# Copyright 2014 Adobe. All rights reserved.

"""
convertfonttocid.py. v 1.13.2 Jul 30 2020

Convert a Type 1 font to CID, given multiple hint dict defs in the
"fontinfo" file. See psautohint help, with the "--doc-fddict" option,
or the makeotf user guide for details on this format. The output file
produced by convertFontToCID() is a CID Type 1 font, no matter what
the input is.

PROCEDURE:
1. convertFontToCID()
   - read 'fontinfo' file
   - getGlyphList(): get list of glyph names (via 'tx -dump -4')
   - getFontBBox(): get FontBBox (via 'tx -mtx -2')
   - getFontName(): get FontName (via 'tx -dump -0')
   - getBlueFuzz(): get BlueFuzz (via 'tx -dump -0')

2. parseFontInfoFile()
   - parse 'fontinfo' file
   - collect all FDDict and corresponding GlyphSet

3. makeSortedGlyphLists()
   - make a list containing a list of glyph names for each FDDict;
     the list is sorted by FDDict index.
     NOTE: all glyphs in a list are contiguous and ordered by original GID,
           and belong to the same FDDict. This allows the glyphs to be
           merged into a single CID font that will have the same glyph
           order as the original font.

4. makeTempFonts()
   - generate a Type1 font (via 'tx -t1 -g ...') for each FDDict;
     the fonts are subsets of the original font.
   - fixFontDict():
     - convert each name-keyed Type1 font to text (via 'detype1')
     - fix FontName
     - fix FontMatrix
     - fix StemSnapH
     - fix StemSnapV
     - fix LanguageGroup
     - fix BlueValues
     - fix OtherBlues
     - convert text font back to Type1 (via 'type1')

5. merge_fonts()
   - makeCIDFontInfo():
     - get the original font's FontDict (via 'tx -0')
     - assemble and save a temporary 'cidfontinfo' file
   - makeGAFile():
     - generate temporary glyph alias file for each FDDict
   - convert each name-keyed Type1 font to CID-keyed (via 'mergefonts')
   - merge all CID-keyed fonts (via 'mergefonts')

"""

import os
import re
import sys

from afdko import fdkutils

__version__ = "1.13.1"

# Tokens seen in font info file that are not part
# of a FDDict or GlyphSet definition.
kBeginToken = "begin"
kEndToken = "end"
kFDDictToken = "FDDict"
kGlyphSetToken = "GlyphSet"

kBaseStateTokens = [
    "FontName",
    "FullName",
    "IsBoldStyle",
    "IsItalicStyle",
    "ConvertToCID",
    "PreferOS/2TypoMetrics",
    "IsOS/2WidthWeigthSlopeOnly",
    "IsOS/2OBLIQUE",
    "UseOldNameID4",
    "LicenseCode"
]

kBlueValueKeys = [
    "BaselineOvershoot",  # 0
    "BaselineYCoord",  # 1
    "CapHeight",  # 2
    "CapOvershoot",  # 3
    "LcHeight",  # 4
    "LcOvershoot",  # 5
    "AscenderHeight",  # 6
    "AscenderOvershoot",  # 7
    "FigHeight",  # 8
    "FigOvershoot",  # 9
    "Height5",  # 10
    "Height5Overshoot",  # 11
    "Height6",  # 12
    "Height6Overshoot",  # 13
]

kOtherBlueValueKeys = [
    "Baseline5Overshoot",  # 0
    "Baseline5",  # 1
    "Baseline6Overshoot",  # 2
    "Baseline6",  # 3
    "SuperiorOvershoot",  # 4
    "SuperiorBaseline",  # 5
    "OrdinalOvershoot",  # 6
    "OrdinalBaseline",  # 7
    "DescenderOvershoot",  # 8
    "DescenderHeight",  # 9
]

kOtherFDDictKeys = [
    "FontName",
    "OrigEmSqUnits",
    "LanguageGroup",
    "DominantV",
    "DominantH",
    "FlexOK",
    "VCounterChars",
    "HCounterChars",
    "BlueFuzz"
]

# We keep this in the FDDict, as it is easier
# to sort and validate as a list of pairs
kFontDictBluePairsName = "BlueValuesPairs"
kFontDictOtherBluePairsName = "OtherBlueValuesPairs"
# Holds the actual string for the Type1 font dict
kFontDictBluesName = "BlueValues"
# Holds the actual string for the Type1 font dict
kFontDictOtherBluesName = "OtherBlues"
kRunTimeFDDictKeys = [
    "DictName",
    kFontDictBluePairsName,
    kFontDictOtherBluePairsName,
    kFontDictBluesName,
    kFontDictOtherBluesName
]
kFDDictKeys = (kOtherFDDictKeys + kBlueValueKeys + kOtherBlueValueKeys +
               kRunTimeFDDictKeys)


# CID stuff
kRequiredCIDFontInfoFields = [
    "FontName",
    "Registry",
    "Ordering",
    "Supplement",
]

kOptionalFields = [
    "FSType",
    "Weight",
    "Trademark",
    "FullName",
    "FamilyName",
    "version",
    "AdobeCopyright",
    "Trademark",
]
kCIDFontInfokeyList = kRequiredCIDFontInfoFields + kOptionalFields

TX_FIELDS = [  # [field name in tx, key name in cidfontinfo, value]
    ["FontName", "FontName", ""],
    ["FullName", "FullName", ""],
    ["FamilyName", "FamilyName", ""],
    ["version", "version", ""],
    ["Notice", "AdobeCopyright", ""],
    ["Copyright", "AdobeCopyright", ""],
    ["Trademark", "Trademark", ""],
    ["FSType", "FSType", ""],
    ["Weight", "Weight", ""],
]


class FontInfoParseError(Exception):
    pass


class FontParseError(Exception):
    pass


class FDDict(object):

    def __init__(self):
        self.DictName = None
        for key in kFDDictKeys:
            setattr(self, key, None)
        self.FlexOK = "true"

    def __repr__(self):
        fddict = {}
        for key, val in vars(self).items():
            if val is None:
                continue
            fddict[key] = val
        return "<%s '%s' %s>" % (
            self.__class__.__name__, fddict.get('DictName', 'no name'), fddict)

    def getFontInfo(self):
        fiList = []
        for key, val in vars(self).items():
            if val is not None:
                fiList.append("%s %s" % (key, val))
        return "\n".join(fiList)

    def buildBlueLists(self):
        baseline_overshoot = getattr(self, 'BaselineOvershoot', None)
        if baseline_overshoot is None:
            print("Error: FDDict definition %s is missing the "
                  "BaselineYCoord/BaselineOvershoot values. These are "
                  "required." % self.DictName)
        elif int(baseline_overshoot) > 0:
            print("Error: The BaselineYCoord/BaselineOvershoot in FDDict "
                  "'%s' must be a bottom zone therefore BaselineOvershoot "
                  "must be negative, not positive." % self.DictName)

        blueKeyList = [kBlueValueKeys, kOtherBlueValueKeys]
        bluePairListNames = [kFontDictBluePairsName,
                             kFontDictOtherBluePairsName]
        blueFieldNames = [kFontDictBluesName, kFontDictOtherBluesName]
        for i in [0, 1]:
            keyList = blueKeyList[i]
            fieldName = blueFieldNames[i]
            pairFieldName = bluePairListNames[i]
            bluePairList = []
            for key in keyList:
                if key.endswith("Overshoot"):
                    width = getattr(self, key)
                    if width is not None:
                        width = int(width)
                        baseName = key[:-len("Overshoot")]
                        if key == "BaselineOvershoot":
                            tempKey = "BaselineYCoord"
                            zonePos = int(getattr(self, tempKey))
                        else:
                            for posSuffix in ["", "Height", "Baseline"]:
                                tempKey = "%s%s" % (baseName, posSuffix)
                                try:
                                    zonePos = int(getattr(self, tempKey))
                                    break
                                except AttributeError:
                                    continue
                        if zonePos is None:
                            raise FontInfoParseError(
                                "Failed to find fontinfo FDDict %s "
                                "top/bottom zone name %s to match the zone "
                                "width key '%s'." % (self.DictName, tempKey,
                                                     key))
                        if width < 0:
                            topPos = zonePos
                            bottomPos = zonePos + width
                            isBottomZone = 1
                            if (i == 0) and (key != "BaselineOvershoot"):
                                print("Error: FontDict %s. Zone %s is a top "
                                      "zone, and the width (%s)  must be "
                                      "positive." % (self.DictName, tempKey,
                                                     width))
                        else:
                            bottomPos = zonePos
                            topPos = zonePos + width
                            isBottomZone = 0
                            if i == 1:
                                print("Error: FontDict %s. Zone %s is a "
                                      "bottom zone, and so the width (%s) "
                                      "must be negative.." % (
                                          self.DictName, tempKey, width))
                        bluePairList.append((topPos, bottomPos, tempKey,
                                             self.DictName, isBottomZone))

            if bluePairList:
                bluePairList.sort()
                prevPair = bluePairList[0]
                zoneBuffer = 2 * getattr(self, 'BlueFuzz', 0) + 1
                for pair in bluePairList[1:]:
                    if prevPair[0] > pair[1]:
                        print("Error in FDDict %s. The top of zone %s at %s "
                              "overlaps zone %s with the bottom at %s." %
                              (self.DictName, prevPair[2], prevPair[0],
                               pair[2], pair[1]))
                    elif abs(pair[1] - prevPair[0]) <= zoneBuffer:
                        print("Error in FDDict %s. The top of zone %s at %s "
                              "is within the min spearation limit (%s units) "
                              "of zone %s with the bottom at %s." %
                              (self.DictName, prevPair[2], prevPair[0],
                               zoneBuffer, pair[2], pair[1]))
                    prevPair = pair
                setattr(self, pairFieldName, bluePairList)
                bluesList = []
                for pairEntry in bluePairList:
                    bluesList.append(pairEntry[1])
                    bluesList.append(pairEntry[0])
                bluesList = [str(val) for val in bluesList]
                bluesList = "[%s]" % (" ".join(bluesList))
                setattr(self, fieldName, bluesList)


def parseFontInfoFile(fontDictList, fontInfoData, glyphList, maxY, minY,
                      fontName, blueFuzz):
    """
    fontDictList may or may not already contain a font dict taken from
    the source font top FontDict.

    Returns fdGlyphDict, fontDictList, finalFDict

    fdGlyphDict: { '.notdef': [0, 0],  # 'g_name': [FDDict_index, g_index]
                   'negative': [1, 2],
                   'a': [2, 1]}

    fontDictList: [<FDDict 'No Alignment Zones' {
                           'FontName': 'SourceSans-Test', 'BlueFuzz': 0,
                           'CapHeight': 760, 'CapOvershoot': 0,
                           'FlexOK': 'true',
                           'BlueValues': '[-112 -112 760 760]',
                           'BlueValuesPairs': [(-112, -112,
                                                'BaselineYCoord',
                                                'No Alignment Zones', 0),
                                               (760, 760,
                                                'CapHeight',
                                                'No Alignment Zones', 0)],
                           'BaselineOvershoot': 0,
                           'DictName': 'No Alignment Zones',
                           'BaselineYCoord': -112}>,
                   <FDDict 'OTHER' {
                           'FontName': 'SourceSans-Test', 'BlueFuzz': 0,
                           'DominantH': '[68]', 'CapHeight': '656',
                           'DominantV': '[86]', 'CapOvershoot': '12',
                           'BlueValues': '[-12 0 656 668]', 'FlexOK': 'false',
                           'BlueValuesPairs': [(0, -12,
                                                'BaselineYCoord','OTHER', 1),
                                               (668, 656,
                                                'CapHeight', 'OTHER', 0)],
                           'BaselineOvershoot': '-12', 'DictName': 'OTHER',
                           'BaselineYCoord': '0'}>,
                   <FDDict 'LOWERCASE' {
                           'FontName': 'SourceSans-Test', 'BlueFuzz': 0,
                           'AscenderHeight': '712', 'DominantH': '[68]',
                           'DescenderOvershoot': '-12', 'DominantV': '[82]',
                           'BlueValues': '[-12 0 486 498 712 724]',
                           'DescenderHeight': '-205', 'LcHeight': '486',
                           'FlexOK': 'false', 'AscenderOvershoot': '12',
                           'LcOvershoot': '12',
                           'BaselineOvershoot': '-12',
                           'OtherBlueValuesPairs': [(-205, -217,
                                                     'DescenderHeight',
                                                     'LOWERCASE', 1)],
                           'BlueValuesPairs': [(0, -12,
                                                'BaselineYCoord',
                                                'LOWERCASE', 1),
                                               (498, 486,
                                                'LcHeight', 'LOWERCASE', 0),
                                               (724, 712,
                                                'AscenderHeight',
                                                'LOWERCASE', 0)],
                           'DictName': 'LOWERCASE',
                           'OtherBlues': '[-217 -205]',
                           'BaselineYCoord': '0'}>]
    """
    # The map of glyph names to font dict: the index into fontDictList.
    fdGlyphDict = {}

    # The user-specified set of blue values to write into the output font,
    # some sort of merge of the individual font dicts. May not be supplied.
    finalFDict = None

    # Get rid of comments.
    data = re.sub(r"#[^\r\n]+[\r\n]", "", fontInfoData)

    # We assume that no items contain whitespace.
    tokenList = data.split()
    numTokens = len(tokenList)
    i = 0
    baseState = 0
    inValue = 1
    inDictValue = 2
    dictState = 3
    glyphSetState = 4
    fdIndexDict = {}
    lenSrcFontDictList = len(fontDictList)
    dictValueList = []
    dictKeyWord = None

    state = baseState

    while i < numTokens:
        token = tokenList[i]
        i += 1
        if state == baseState:
            if token == kBeginToken:
                token = tokenList[i]
                i += 1
                if token == kFDDictToken:
                    state = dictState
                    dictName = tokenList[i]
                    i += 1
                    fdDict = FDDict()
                    fdDict.DictName = dictName
                    if dictName == "FinalFont":
                        # This dict is NOT used to hint any glyphs;
                        # it is used to supply the merged alignment zones
                        # and stem widths for the final font.
                        finalFDict = fdDict
                    else:
                        # save dict and FDIndex
                        fdIndexDict[dictName] = len(fontDictList)
                        fontDictList.append(fdDict)

                elif token == kGlyphSetToken:
                    state = glyphSetState
                    setName = tokenList[i]
                    i += 1
                else:
                    raise FontInfoParseError(
                        "Unrecognized token after \"begin\" keyword: %s" %
                        token)

            elif token in kBaseStateTokens:
                # Discard value for base token.
                token = tokenList[i]
                i += 1
                if (token[0] in ["[", "("]) and (token[-1] not in ["]", ")"]):
                    state = inValue
            else:
                raise FontInfoParseError(
                    "Unrecognized token in base state: %s" % token)

        elif state == inValue:
            # We are processing a list value for a base state token.
            if token[-1] in ["]", ")"]:
                # found the last token in the list value.
                state = baseState

        elif state == inDictValue:
            dictValueList.append(token)
            if token[-1] in ["]", ")"]:
                value = " ".join(dictValueList)
                setattr(fdDict, dictKeyWord, value)
                # found the last token in the list value.
                state = dictState

        elif state == glyphSetState:
            # "end GlyphSet" marks end of set,
            # else we are adding a new glyph name.
            if (token == kEndToken) and tokenList[i] == kGlyphSetToken:
                if tokenList[i + 1] != setName:
                    raise FontInfoParseError(
                        "End glyph set name \"%s\" does not match begin "
                        "glyph set name \"%s\"." % (tokenList[i + 1], setName))
                state = baseState
                i += 2
                setName = None
            else:
                # Need to add matching glyphs.
                gi = 0
                for gname in glyphList:
                    if re.search(token, gname):
                        # fdIndex value
                        fdGlyphDict[gname] = [fdIndexDict[setName], gi]
                    gi += 1

        elif state == dictState:
            # "end FDDict" marks end of set,
            # else we are adding a new glyph name.
            if (token == kEndToken) and tokenList[i] == kFDDictToken:
                if tokenList[i + 1] != dictName:
                    raise FontInfoParseError(
                        "End FDDict  name \"%s\" does not match begin FDDict "
                        "name \"%s\"." % (tokenList[i + 1], dictName))
                if fdDict.DominantH is None:
                    print("Warning: the FDDict '%s' in fontinfo has no "
                          "DominantH value" % dictName)
                if fdDict.DominantV is None:
                    print("Warning: the FDDict '%s' in fontinfo has no "
                          "DominantV value" % dictName)
                if fdDict.BlueFuzz is None:
                    fdDict.BlueFuzz = blueFuzz
                fdDict.buildBlueLists()
                if fdDict.FontName is None:
                    fdDict.FontName = fontName
                state = baseState
                i += 2
                dictName = None
                fdDict = None
            else:
                if token in kFDDictKeys:
                    value = tokenList[i]
                    i += 1
                    if ((value[0] in ["[", "("]) and
                       (value[-1] not in ["]", ")"])):
                        state = inDictValue
                        dictValueList = [value]
                        dictKeyWord = token
                    else:
                        setattr(fdDict, token, value)
                else:
                    raise FontInfoParseError(
                        "FDDict key \"%s\" in fdDict named \"%s\" is not "
                        "recognized." % (token, dictName))

    if lenSrcFontDictList != len(fontDictList):
        # There are some FDDict definitions. This means that we need
        # to fix the default fontDict, inherited from the source font,
        # so that it has blues zones that will not affect hinting,
        # e.g outside of the Font BBox. We do this because if there are
        # glyphs which are not assigned to a user specified font dict,
        # it is because it doesn't make sense to provide alignment zones
        # for the glyph. Since psautohint does require at least one bottom zone
        # and one top zone, we add one bottom and one top zone that are
        # outside the font BBox, so that hinting won't be affected by them.
        # NOTE: The FDDict receives a special name "No Alignment Zones" which
        # psautohint recognizes.
        defaultFDDict = fontDictList[0]

        for key in kBlueValueKeys + kOtherBlueValueKeys:
            setattr(defaultFDDict, key, None)

        defaultFDDict.BaselineYCoord = minY - 100
        defaultFDDict.BaselineOvershoot = 0
        defaultFDDict.CapHeight = maxY + 100
        defaultFDDict.CapOvershoot = 0
        defaultFDDict.BlueFuzz = 0
        defaultFDDict.DictName = "No Alignment Zones"
        defaultFDDict.FontName = fontName
        defaultFDDict.buildBlueLists()

        for gi, gname in enumerate(glyphList):
            if gname not in fdGlyphDict:
                fdGlyphDict[gname] = [0, gi]

    # hard code the '.notdef' to the default FDDict;
    # this is only necessary when fdGlyphDict is not empty.
    # NOTE: at this point it's guaranteed that the font has a glyph
    #       named '.notdef' at GID 0.
    #       If the '.notdef' was included in one of the GlyphSet definitions
    #       it will be associated with an FDDict other than the default.
    if fdGlyphDict:
        fdGlyphDict['.notdef'] = [0, 0]

    return fdGlyphDict, fontDictList, finalFDict


def mergeFDDicts(prevDictList, privateDict):
    """
    Used by beztools & ufotools.
    """
    # Extract the union of the stem widths and zones from the list
    # of FDDicts, and replace the current values in the topDict.
    fdDictName = None
    prefDDict = None
    zoneName = None
    blueZoneDict = {}
    otherBlueZoneDict = {}
    dominantHDict = {}
    dominantVDict = {}
    bluePairListNames = [kFontDictBluePairsName, kFontDictOtherBluePairsName]
    zoneDictList = [blueZoneDict, otherBlueZoneDict]
    for prefDDict in prevDictList:
        for ki in [0, 1]:
            zoneDict = zoneDictList[ki]
            bluePairName = bluePairListNames[ki]
            bluePairList = getattr(prefDDict, bluePairName)
            if not bluePairList:
                continue
            for topPos, bottomPos, zoneName, _, isBottomZone in bluePairList:
                zoneDict[(topPos, bottomPos)] = (isBottomZone, zoneName,
                                                 prefDDict.DictName)

        # Now for the stem widths.
        stemNameList = ["DominantH", "DominantV"]
        stemDictList = [dominantHDict, dominantVDict]
        for wi in (0, 1):
            stemFieldName = stemNameList[wi]
            dList = getattr(prefDDict, stemFieldName)
            stemDict = stemDictList[wi]
            if dList is not None:
                dList = dList[1:-1]  # remove the braces
                dList = dList.split()
                dList = [int(val) for val in dList]
                for width in dList:
                    stemDict[width] = prefDDict.DictName

    # Now we have collected all the stem widths and zones
    # from all the dicts. See if we can merge them.
    goodBlueZoneList = []
    goodOtherBlueZoneList = []
    goodHStemList = []
    goodVStemList = []

    zoneDictList = [blueZoneDict, otherBlueZoneDict]
    goodZoneLists = [goodBlueZoneList, goodOtherBlueZoneList]
    stemDictList = [dominantHDict, dominantVDict]
    goodStemLists = [goodHStemList, goodVStemList]

    for ki in [0, 1]:
        zoneDict = zoneDictList[ki]
        goodZoneList = goodZoneLists[ki]
        stemDict = stemDictList[ki]
        goodStemList = goodStemLists[ki]

        # Zones first.
        zoneList = sorted(zoneDict.keys())
        if not zoneList:
            continue

        # Now check for conflicts.
        prevZone = zoneList[0]
        goodZoneList.append(prevZone[1])
        goodZoneList.append(prevZone[0])
        zoneBuffer = 2 * prefDDict.BlueFuzz + 1
        for zone in zoneList[1:]:
            if (ki == 0) and (len(zoneList) >= 14):
                print("Warning. For final FontDict, skipping BlueValues "
                      "alignment zone %s from FDDict %s because there are "
                      "already 7 zones." % (zoneName, fdDictName))
            elif (ki == 1) and (len(zoneList) >= 5):
                print("Warning. For final FontDict, skipping OtherBlues "
                      "alignment zone %s from FDDict %s because there are "
                      "already 5 zones." % (zoneName, fdDictName))
            if zone[1] < prevZone[0]:
                curEntry = blueZoneDict[zone]
                prevEntry = blueZoneDict[prevZone]
                zoneName = curEntry[1]
                fdDictName = curEntry[2]
                prevZoneName = prevEntry[1]
                prevFDictName = prevEntry[2]
                print("Warning. For final FontDict, skipping zone %s in "
                      "FDDict %s because it overlaps with zone %s in "
                      "FDDict %s." % (zoneName, fdDictName, prevZoneName,
                                      prevFDictName))
            elif abs(zone[1] - prevZone[0]) <= zoneBuffer:
                curEntry = blueZoneDict[zone]
                prevEntry = blueZoneDict[prevZone]
                zoneName = curEntry[1]
                fdDictName = curEntry[2]
                prevZoneName = prevEntry[1]
                prevFDictName = prevEntry[2]
                print("Warning. For final FontDict, skipping zone %s in "
                      "FDDict %s because it is within the minimum separation "
                      "allowed (%s units) of %s in FDDict %s." %
                      (zoneName, fdDictName, zoneBuffer, prevZoneName,
                       prevFDictName))
            else:
                goodZoneList.append(zone[1])
                goodZoneList.append(zone[0])
            prevZone = zone

        stemList = sorted(stemDict.keys())
        if not stemList:
            continue

        # Now check for conflicts.
        prevStem = stemList[0]
        goodStemList.append(prevStem)
        for stem in stemList[1:]:
            if abs(stem - prevStem) < 2:
                fdDictName = stemDict[stem]
                prevFDictName = stemDict[prevStem]
                print("Warning. For final FontDict, skipping stem width %s "
                      "in FDDict %s because it overlaps in coverage with "
                      "stem width %s in FDDict %s." %
                      (stem, fdDictName, prevStem, prevFDictName))
            else:
                goodStemList.append(stem)
            prevStem = stem

    if goodBlueZoneList:
        privateDict.BlueValues = goodBlueZoneList

    if goodOtherBlueZoneList:
        privateDict.OtherBlues = goodOtherBlueZoneList
    else:
        privateDict.OtherBlues = None

    if goodHStemList:
        privateDict.StemSnapH = goodHStemList
    else:
        privateDict.StemSnapH = None

    if goodVStemList:
        privateDict.StemSnapV = goodVStemList
    else:
        privateDict.StemSnapV = None


def getGlyphList(fPath, removeNotdef=False, original_font=False):
    command = "tx -dump -4 \"%s\" 2>&1" % fPath
    data = fdkutils.runShellCmd(command)
    if not data:
        raise FontParseError("Error: Failed running 'tx -dump -4' on file "
                             "%s" % fPath)

    # initial newline keeps us from picking up the first column header line
    nameList = re.findall(r"[\r\n]glyph.+?{([^,]+),", data)
    if not nameList:
        raise FontParseError("Error: Failed getting glyph names from file %s "
                             "using tx." % fPath)

    if original_font:
        # make sure the original font has a .notdef glyph
        # and that it's the first glyph
        if '.notdef' not in nameList:
            raise FontParseError("Error: The font file %s does not have a "
                                 "'.notdef'." % fPath)
        elif nameList[0] != '.notdef':
            raise FontParseError("Error: '.notdef' is not the first glyph of "
                                 "font file %s" % fPath)

    if removeNotdef:
        nameList.remove(".notdef")

    return nameList


def getFontBBox(fPath):
    command = "tx -mtx -2  \"%s\" 2>&1" % fPath
    data = fdkutils.runShellCmd(command)
    if not data:
        raise FontInfoParseError("Error: Failed getting log from tx from %s, "
                                 "when trying to get FontBBox." % fPath)

    m = re.search(r"bbox[^{]+{([^}]+)}", data)
    if not m:
        print(data)
        raise FontInfoParseError("Error: Failed finding FontBBox in tx log "
                                 "from %s." % fPath)
    fontBBox = m.group(1).split(",")
    fontBBox = [int(val) for val in fontBBox]
    return fontBBox


def getFontName(fPath):
    command = "tx -dump -0 \"%s\" 2>&1" % fPath
    data = fdkutils.runShellCmd(command)
    if not data:
        raise FontInfoParseError("Error: Failed getting log from tx from %s, "
                                 "when trying to get FontName." % fPath)

    m = re.search(r"FontName\s+\"([^\"]+)", data)
    if not m:
        print(data)
        raise FontInfoParseError("Error: Failed finding FontName in tx log "
                                 "from %s." % fPath)
    return m.group(1)


def getBlueFuzz(fPath):
    blueFuzz = 1.0
    command = "tx -dump -0 \"%s\" 2>&1" % fPath
    data = fdkutils.runShellCmd(command)
    if not data:
        raise FontInfoParseError("Error: Failed getting log from tx from %s, "
                                 "when trying to get FontName." % fPath)

    m = re.search(r"BlueFuzz\s+(\d+(?:\.\d+)*)", data)
    if m:
        blueFuzz = int(m.group(1))
    return blueFuzz


def makeSortedGlyphLists(glyphList, fdGlyphDict):
    """
    Returns a list containing lists of glyph names (one for each FDDict).

    glyphList: list of all glyph names in the font

    fdGlyphDict: {'a': [2, 1], 'negative': [1, 2], '.notdef': [0, 0]}
                 keys: glyph names
                 values: [FDDict_index, glyph_index]
    """
    glyphSetList = [[]]
    for glyph_name in glyphList:
        try:
            fddict_index = fdGlyphDict[glyph_name][0]
            while len(glyphSetList) <= fddict_index:
                # glyphSetList does not have enough empty containers/lists
                glyphSetList.append([])
            glyphSetList[fddict_index].append(glyph_name)
        except KeyError:
            # the glyph name is not in the FDDict;
            # assign that glyph to the default FDDict (i.e. index 0)
            glyphSetList[0].append(glyph_name)

    if len(glyphSetList) < 2:
        print("Warning: There is only one hint dict. Maybe the fontinfo file "
              "has not been edited to add hint dict info.")

    return glyphSetList


def fixFontDict(tempPath, fdDict):
    txtPath = fdkutils.get_temp_file_path()
    command = "detype1 \"%s\" \"%s\" 2>&1" % (tempPath, txtPath)
    log = fdkutils.runShellCmd(command)
    if log:
        print(log)

    with open(txtPath, "r", encoding='utf-8') as fp:
        data = fp.read()

    # fix font name. We always search for it, as it is always present,
    # and we can use the following white space to get the file new line.
    m = re.search(r"(/FontName\s+/\S+\s+def)(\s+)", data)
    newLine = m.group(2)
    if not m:
        raise FontParseError("Failed to find FontName in input font! "
                             "%s" % tempPath)
    if fdDict.FontName:
        target = "/FontName /%s def" % fdDict.FontName
        data = data[:m.start(1)] + target + data[m.end(1):]

    # fix em square
    if fdDict.OrigEmSqUnits:
        m = re.search(r"/FontMatrix\s+\[.+?\]\s+def", data)
        if not m:
            raise FontParseError("Failed to find FontMatrix in input font! "
                                 "%s" % tempPath)
        emUnits = fdDict.OrigEmSqUnits
        a = 1.0 / emUnits
        target = "/FontMatrix [%s 0 0 %s 0 0] def" % (a, a)
        data = data[:m.start()] + target + data[m.end():]

    # fix StemSnapH. Remove StemSnapH if fdDict.StemSnapH
    # is not defined, else set it.
    m = re.search(r"/StemSnapH\s+\[.+?\]\s+def", data)
    if fdDict.DominantH:
        target = "/StemSnapH %s def" % fdDict.DominantH
        data = data[:m.start()] + target + data[m.end():]
        insertIndex = m.start() + len(target)
    else:
        if m:
            data = data[:m.start()] + data[m.end():]

    # fix StemSnapV.  Remove StemSnapV entry if fdDict.StemSnapV
    # is not defined, else set it.
    m = re.search(r"/StemSnapV\s+\[.+?\]\s+def", data)
    if fdDict.DominantV:
        target = "/StemSnapV %s def" % fdDict.DominantV
        data = data[:m.start()] + target + data[m.end():]
        insertIndex = m.start() + len(target)
    else:
        if m:
            data = data[:m.start()] + data[m.end():]

    # LanguageGroup. Remove LanguageGroup entry if
    # fdDict.LanguageGroup is not defined, else set it.
    if fdDict.LanguageGroup:
        m = re.search(r"/LanguageGroup\s+\d+\s+def", data)
        if not m:
            target = "%s/LanguageGroup %s def" % (newLine,
                                                  fdDict.LanguageGroup)
            data = data[:insertIndex] + data[insertIndex] + target + \
                data[insertIndex:]
        else:
            target = "/LanguageGroup %s def" % fdDict.LanguageGroup
            data = data[:m.start()] + target + data[m.end():]
    else:
        m = re.search(r"/LanguageGroup\s+\d+\s+def", data)
        if m:
            data = data[:m.start()] + data[m.end():]

    # Fix BlueValues. Must be present.
    if fdDict.BlueValues:
        m = re.search(r"/BlueValues\s+\[.+?\]\s+def", data)
        if not m:
            raise FontParseError("Failed to find BlueValues in input font! "
                                 "%s" % tempPath)
        target = "/BlueValues %s def" % fdDict.BlueValues
        data = data[:m.start()] + target + data[m.end():]
        insertIndex = m.start() + len(target)

    # Fix OtherBlues, if present. Remove if there are no OtherBlues entry.
    m = re.search(r"/OtherBlues\s+\[.+?\]\s+def", data)
    if fdDict.OtherBlues:
        if not m:
            target = "%s/OtherBlues %s def" % (newLine, fdDict.OtherBlues)
            data = data[:insertIndex] + target + data[insertIndex:]
        else:
            target = "/OtherBlues %s def" % fdDict.OtherBlues
            data = data[:m.start()] + target + data[m.end():]
    else:
        if m:
            data = data[:m.start()] + data[m.end():]

    with open(txtPath, "w") as fp:
        fp.write(data)

    command = "type1 \"%s\" \"%s\" 2>&1" % (txtPath, tempPath)
    log = fdkutils.runShellCmd(command)
    if log:
        print(log)


def makeTempFonts(fontDictList, glyphSetList, fdGlyphDict, inputPath):
    fontList = []
    for glyphList in glyphSetList:
        arg = ",".join(glyphList)
        tempPath = fdkutils.get_temp_file_path()
        command = "tx -t1 -g \"%s\" \"%s\" \"%s\" 2>&1" % (
            arg, inputPath, tempPath)
        log = fdkutils.runShellCmd(command)
        if log:
            print("Have log output in subsetting command for %s to %s with "
                  "%s glyphs." % (inputPath, tempPath, len(glyphList)))
            print(log)

        try:
            fdIndex = fdGlyphDict[glyphList[0]][0]
        except KeyError:
            # 'fontinfo' file was not provided;
            # assign list of glyphs to default FDDict
            fdIndex = 0

        fdDict = fontDictList[fdIndex]
        fixFontDict(tempPath, fdDict)
        fontList.append(tempPath)

    return fontList


def makeCIDFontInfo(fontPath, cidfontinfoPath):
    cfiDict = {}
    for key in kCIDFontInfokeyList:
        cfiDict[key] = None
    cfiDict["Weight"] = "(Regular)"
    cfiDict["AdobeCopyright"] = "0"

    # get regular FontDict.
    command = "tx -0 \"%s\" 2>&1" % fontPath
    report = fdkutils.runShellCmd(command)
    if ("fatal" in report) or ("error" in report):
        print(report)
        raise FontInfoParseError("Failed to dump font dict using tx from "
                                 "font '%s'" % fontPath)

    for entry in TX_FIELDS:
        match = re.search(entry[0] + r"\s+(.+?)[\r\n]", report)
        if match:
            entry[2] = match.group(1)

    cfiDict["Registry"] = "Adobe"
    cfiDict["Ordering"] = "Identity"
    cfiDict["Supplement"] = "0"

    for entry in TX_FIELDS:
        if entry[2]:
            cfiDict[entry[1]] = entry[2]
        elif entry[1] in kRequiredCIDFontInfoFields:
            print("Error: did not find required info '%s' in tx dump of "
                  "font '%s'." % (entry[1], fontPath))
    try:
        with open(cidfontinfoPath, "w") as fp:
            for key in kCIDFontInfokeyList:
                value = cfiDict[key]
                if value is None:
                    continue
                if value[0] == "\"":
                    value = "(" + value[1:-1] + ")"
                string = "%s\t%s\n" % (key, value)
                fp.write(string)
    except (OSError):
        raise FontInfoParseError(
            "Error. Could not open and write file '%s'" % cidfontinfoPath)


def makeGAFile(gaPath, fontPath, glyphList, fontDictList, fdGlyphDict,
               removeNotdef):
    """
    Creates a glyph alias file for each FDDict.
    These files will be used by 'mergefonts' tool.
    For documentation on the format of this file, run 'mergefonts -h'.
    """
    glyph_list = getGlyphList(fontPath, removeNotdef)

    try:
        fdIndex = fdGlyphDict[glyph_list[0]][0]  # [fdIndex value, gi]
    except KeyError:
        fdIndex = 0

    fdDict = fontDictList[fdIndex]
    lineList = [""]

    lang_group = fdDict.LanguageGroup
    if lang_group is None:
        langGroup = " 0"
    else:
        langGroup = " %s" % lang_group

    dictName = "%s_%s" % (fdDict.FontName, fdDict.DictName)

    for glyph_name in glyph_list:
        gid = glyphList.index(glyph_name)
        lineList.append("%s\t%s" % (gid, glyph_name))

    lineList.append("")
    gaText = "mergefonts %s%s%s" % (dictName, langGroup, '\n'.join(lineList))

    with open(gaPath, "w") as gf:
        gf.write(gaText)


def merge_fonts(inputFontPath, outputPath, fontList, glyphList, fontDictList,
                fdGlyphDict):
    cidfontinfoPath = fdkutils.get_temp_file_path()
    makeCIDFontInfo(inputFontPath, cidfontinfoPath)
    lastFont = ""
    dstPath = ''

    for i, fontPath in enumerate(fontList):
        gaPath = fdkutils.get_temp_file_path()
        dstPath = fdkutils.get_temp_file_path()
        removeNotdef = i != 0
        makeGAFile(gaPath, fontPath, glyphList, fontDictList, fdGlyphDict,
                   removeNotdef)
        if lastFont:
            command = 'mergefonts -std -cid "%s" "%s" "%s" "%s" "%s" 2>&1' % (
                cidfontinfoPath, dstPath, lastFont, gaPath, fontPath)
        else:
            command = 'mergefonts -std -cid "%s" "%s" "%s" "%s" 2>&1' % (
                cidfontinfoPath, dstPath, gaPath, fontPath)
        log = fdkutils.runShellCmd(command)
        if "rror" in log:
            raise FontInfoParseError(
                "Error running command '%s'\nLog: %s" % (command, log))

        lastFont = dstPath

    if os.path.exists(outputPath):
        os.remove(outputPath)
        os.rename(dstPath, outputPath)


def convertFontToCID(inputPath, outputPath, fontinfoPath=None):
    """
    Takes in a path to the font file to convert, a path to save the result,
    and an optional path to a '(cid)fontinfo' file.
    """
    if fontinfoPath and os.path.exists(fontinfoPath):
        with open(fontinfoPath, "r", encoding='utf-8') as fi:
            fontInfoData = fi.read()
    else:
        fontInfoData = ''

    glyphList = getGlyphList(inputPath, False, True)
    fontBBox = getFontBBox(inputPath)
    fontName = getFontName(inputPath)
    blueFuzz = getBlueFuzz(inputPath)
    maxY = fontBBox[3]
    minY = fontBBox[1]

    # Start with an uninitialized entry for the default FDDict 0
    fontDictList = [FDDict()]

    fdGlyphDict, fontDictList, _ = parseFontInfoFile(
        fontDictList, fontInfoData, glyphList, maxY, minY, fontName, blueFuzz)

    glyphSetList = makeSortedGlyphLists(glyphList, fdGlyphDict)

    fontList = makeTempFonts(fontDictList, glyphSetList, fdGlyphDict,
                             inputPath)

    merge_fonts(inputPath, outputPath, fontList, glyphList, fontDictList,
                fdGlyphDict)


def mergeFontToCFF(srcPath, outputPath, doSubr):
    """
    Used by makeotf.
    Assumes srcPath is a type 1 font,and outputPath is an OTF font.
    """
    # First, convert src font to cff, and subroutinize it if so requested.
    tempPath = fdkutils.get_temp_file_path()
    subrArg = ""
    if doSubr:
        subrArg = " +S"
    command = "tx -cff +b%s \"%s\" \"%s\" 2>&1" % (subrArg, srcPath, tempPath)
    report = fdkutils.runShellCmd(command)
    if ("fatal" in report) or ("error" in report):
        print(report)
        raise FontInfoParseError(
            "Failed to run 'tx -cff +b' on file %s" % srcPath)

    # Now merge it into the output file.
    command = "sfntedit -a CFF=\"%s\" \"%s\" 2>&1" % (tempPath, outputPath)
    report = fdkutils.runShellCmd(command)
    if ("fatal" in report) or ("error" in report):
        print(report)
        raise FontInfoParseError(
            "Failed to run 'sfntedit -a CFF=' on file %s" % srcPath)


def main():
    args = sys.argv

    if len(args) >= 2:
        inputPath = args[1]
        outputPath = fdkutils.get_temp_file_path()

        fontinfoPath = None
        if len(args) == 3:
            fontinfoPath = args[2]

        convertFontToCID(inputPath, outputPath, fontinfoPath)

        save_path = f'{os.path.splitext(inputPath)[0]}-CID.ps'

        if os.path.isfile(outputPath):
            os.rename(outputPath, save_path)
    else:
        print('ERROR: Missing path to font to convert.')


if __name__ == '__main__':
    main()



# Copyright 2018 Adobe. All rights reserved.

"""
Helper script for running a command line tool.
Used as part of the integration tests.
"""

import argparse
import logging
import os
import subprocess
import sys
import tempfile

__version__ = '0.6.3'

logger = logging.getLogger('runner')


TIMEOUT = 240  # seconds


def _write_file(file_path, data):
    with open(file_path, "wb") as f:
        f.write(data)


def _get_input_dir_path(tool_name):
    input_dir = os.path.join(
        os.path.split(__file__)[0], f'{tool_name}_data', 'input')
    return os.path.abspath(os.path.realpath(input_dir))


def run_tool(opts):
    """
    Runs the tool using the parameters provided.
    """
    input_dir = _get_input_dir_path(opts.tool)

    args = [opts.tool]
    for opt in opts.options:
        if opt.startswith('='):
            args.append(f'--{opt[1:]}')
        elif opt.startswith('*'):
            args.append(f'+{opt[1:]}')
        elif opt.startswith('_'):
            args.append(f'{opt[1:]}')
        else:
            args.append(f'-{opt}')

    if opts.files:
        if opts.abs_paths:
            files = opts.files
            # validate only the first path; the other paths may not exist yet
            # as they can be the paths to which the user wants the tool to
            # write its output to
            assert os.path.exists(files[0]), "Invalid input path found."
        else:
            files = [os.path.join(input_dir, fname) for fname in opts.files]
            assert all([os.path.exists(fpath) for fpath in files]), (
                "Invalid input path found.")
        args.extend(files)

    stderr = None
    if opts.std_error:
        stderr = subprocess.STDOUT

    logger.debug(
        f"About to run the command below\n==>{' '.join(args)}<==")
    try:
        if opts.save_path:
            output = subprocess.check_output(args, stderr=stderr,
                                             timeout=TIMEOUT)
            _write_file(opts.save_path, output)
            return opts.save_path
        else:
            return subprocess.check_call(args, timeout=TIMEOUT)
    except (subprocess.CalledProcessError, OSError) as err:
        if opts.save_path:
            _write_file(opts.save_path, err.output)
            return opts.save_path
        logger.error(err)
        raise


def _check_tool(tool_name):
    """
    Checks if the invoked tool produces a known command.
    Returns the tool's name if the check is successful,
    or a tuple with the tool's name and the error if it's not.
    """
    try:
        subprocess.check_output([tool_name, '-h'], timeout=TIMEOUT)
        return tool_name
    except (subprocess.CalledProcessError, OSError) as err:
        logger.error(err)
        return tool_name, err


def _check_save_path(path_str):
    check_path = os.path.abspath(os.path.realpath(path_str))
    del_test_file = True
    try:
        if os.path.exists(check_path):
            del_test_file = False
        open(check_path, 'a').close()
        if del_test_file:
            os.remove(check_path)
    except (OSError):
        raise argparse.ArgumentTypeError(
            f"{check_path} is not a valid path to write to.")
    return check_path


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='verbose mode\n'
             'Use -vv for debug mode'
    )
    parser.add_argument(
        '-t',
        '--tool',
        required=True,
        type=_check_tool,
        metavar='NAME',
        help='name of the tool to run'
    )
    parser.add_argument(
        '-o',
        '--options',
        nargs='+',
        metavar='OPTION',
        default=[],
        help='options to run the tool with\n'
             'By default the option will be prepended with one minus (-).\n'
             'To request the use of two minus (--), prepend the option with '
             'an equal (=).\n'
             'To request the use of a plus (+), prepend the option with an '
             'asterisk (*).\n'
             'To request the use of nothing, prepend the option with an '
             'underscore (_).'
    )
    parser.add_argument(
        '-f',
        '--files',
        nargs='+',
        metavar='FILE_NAME/PATH',
        help='names or full paths of the files to provide to the tool\n'
             "The files will be sourced from the 'input' folder inside the\n"
             "'{tool_name}_data' directory, unless the option '--abs-paths' "
             "is used."
    )
    parser.add_argument(
        '-a',
        '--abs-paths',
        action='store_true',
        help="treat the input paths as absolute\n"
             "(instead of deriving them from the tool's name)"
    )
    parser.add_argument(
        '-s',
        '--save',
        dest='save_path',
        nargs='?',
        default=False,
        type=_check_save_path,
        metavar='blank or PATH',
        help="path to save the tool's standard output (stdout) to\n"
             'The default is to use a temporary location.'
    )
    parser.add_argument(
        '-e',
        '--std-error',
        action='store_true',
        help="capture stderr instead of stdout"
    )
    options = parser.parse_args(args)

    if not options.verbose:
        level = "WARNING"
    elif options.verbose == 1:
        level = "INFO"
    else:
        level = "DEBUG"
    logging.basicConfig(level=level)

    if options.save_path is None:
        # runner.py was called with '-s' but the option is NOT followed by
        # a command-line argument; this means a temp file must be created.
        # When option '-s' is NOT used, the value of options.save_path is False
        file_descriptor, temp_path = tempfile.mkstemp()
        os.close(file_descriptor)
        options.save_path = temp_path

    if isinstance(options.tool, tuple):
        parser.error(f"'{options.tool[0]}' is an unknown command.")

    return options


def main(args=None):
    """
    Returns the path to the result/output file if the command is successful,
    and an exception if it isn't.
    """
    opts = get_options(args)

    return run_tool(opts)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2015 Adobe. All rights reserved.

"""
Tool that performs outline quality checks and can remove path overlaps.
"""

__version__ = '2.5.0'

import argparse
from functools import cmp_to_key
import re
import shutil
import sys
import textwrap
from tqdm import tqdm, trange
import warnings

import booleanOperations.booleanGlyph
import defcon
from fontPens.digestPointPen import DigestPointPen
from fontTools.ufoLib import UFOWriter, UFOLibError, UFOFormatVersion

from afdko import ufotools
from afdko.ufotools import (
    thresholdAttrGlyph,
    kProcessedGlyphsLayer as PROCD_GLYPHS_LAYER,
    kProcessedGlyphsLayerName as PROCD_GLYPHS_LAYER_NAME,
)
from afdko.fdkutils import (
    get_temp_file_path,
    get_temp_dir_path,
    get_font_format,
    run_shell_command,
    validate_path,
)


UFO_FONT_TYPE = 1
TYPE1_FONT_TYPE = 2
CFF_FONT_TYPE = 3
OTF_FONT_TYPE = 4


class FocusOptionParseError(Exception):
    pass


class FocusFontError(Exception):
    pass


class FontFile(object):
    def __init__(self, font_path, font_format):
        self.font_path = font_path
        self.font_format = font_format
        self.temp_ufo_path = None
        self.font_type = None
        self.defcon_font = None
        self.use_hash_map = False
        self.ufo_font_hash_data = None
        self.save_to_default_layer = False

    def open(self, use_hash_map):
        font_path = self.font_path

        if self.font_format == 'UFO':
            self.font_type = UFO_FONT_TYPE
            ufotools.validateLayers(font_path)
            self.defcon_font = defcon.Font(font_path)
            self.ufo_format = self.defcon_font.ufoFormatVersionTuple
            self.use_hash_map = use_hash_map
            self.ufo_font_hash_data = ufotools.UFOFontData(
                font_path, self.use_hash_map,
                programName=ufotools.kCheckOutlineName)
            self.ufo_font_hash_data.readHashMap()

        else:
            print("Converting to temp UFO font...")
            self.temp_ufo_path = temp_path = get_temp_dir_path('font.ufo')

            if not run_shell_command([
                    'tx', '-ufo', font_path, temp_path]):
                raise FocusFontError(
                    'Failed to convert input font to UFO.')

            try:
                self.defcon_font = defcon.Font(temp_path)
            except UFOLibError:
                raise

            if self.font_format == 'OTF':
                self.font_type = OTF_FONT_TYPE
            elif self.font_format == 'CFF':
                self.font_type = CFF_FONT_TYPE
            else:
                self.font_type = TYPE1_FONT_TYPE

        return self.defcon_font

    def close(self):
        # Differs from save in that it saves just the hash file.
        # To be used when the glif files have not been changed by this program,
        # but the hash file has changed.
        if self.font_type == UFO_FONT_TYPE:
            # Write the hash data, if it has changed.
            self.ufo_font_hash_data.close()

    def save(self):
        if self.save_to_default_layer:
            self.defcon_font.save()
        else:
            if self.ufo_format <= UFOFormatVersion.FORMAT_2_0:
                # Up-convert the UFO to format 3
                warnings.warn("The UFO was up-converted to format 3.")
                self.ufo_format = UFOFormatVersion.FORMAT_3_0

                with UFOWriter(self.defcon_font.path,
                               formatVersion=self.ufo_format) as writer:
                    writer.getGlyphSet()
                    writer.writeLayerContents()

            writer = UFOWriter(
                self.defcon_font.path, formatVersion=self.ufo_format)
            writer.layerContents[
                PROCD_GLYPHS_LAYER_NAME] = PROCD_GLYPHS_LAYER
            layers = self.defcon_font.layers
            layer = layers[PROCD_GLYPHS_LAYER_NAME]
            glyph_set = writer.getGlyphSet(
                layerName=PROCD_GLYPHS_LAYER_NAME, defaultLayer=False)
            writer.writeLayerContents(layers.layerOrder)
            layer.save(glyph_set)

        if self.font_type == UFO_FONT_TYPE:
            ufotools.regenerate_glyph_hashes(self.ufo_font_hash_data)
            # Write the hash data, if it has changed.
            self.ufo_font_hash_data.close()

        elif self.font_type == TYPE1_FONT_TYPE:
            args = ['tx', '-t1']
            if self.font_format == 'PFB':
                args.append('-pfb')
            if not run_shell_command(
                    args + [self.temp_ufo_path, self.font_path]):
                raise FocusFontError('Failed to convert UFO font to Type 1.')

        else:
            temp_cff_path = get_temp_file_path()
            if not run_shell_command([
                    'tx', '-cff', '+S', '+b', '-std',
                    self.temp_ufo_path, temp_cff_path], suppress_output=True):
                raise FocusFontError('Failed to convert UFO font to CFF.')

            if self.font_type == CFF_FONT_TYPE:
                shutil.copy2(temp_cff_path, self.font_path)

            else:  # OTF_FONT_TYPE
                if not run_shell_command([
                        'sfntedit', '-a',
                        f'CFF={temp_cff_path}', self.font_path]):
                    raise FocusFontError('Failed to add CFF table to OTF.')

    def check_skip_glyph(self, glyph_name, do_all):
        skip = False
        if self.ufo_font_hash_data and self.use_hash_map:
            _, skip = \
                self.ufo_font_hash_data.getOrSkipGlyph(glyph_name, do_all)
        return skip

    def clear_hash_map(self):
        if self.ufo_font_hash_data is not None:
            self.ufo_font_hash_data.clearHashMap()


class COOptions(object):
    def __init__(self):
        self.file_path = None
        self.out_file_path = None
        self.log_file_path = None
        self.glyph_list = []
        self.allow_changes = False
        self.write_to_default_layer = False
        self.allow_decimal_coords = False
        self.min_area = 25
        self.tolerance = 0

        # forces all glyphs to be processed even if src hasn't changed.
        self.check_all = False

        # processing state flag, used to not repeat coincident point removal.
        self.remove_coincident_points_done = False

        # processing state flag, used to not repeat flat curve point removal.
        self.remove_flat_curves_done = False

        self.clear_hash_map = False
        self.quiet_mode = False

        # do_overlap_removal must come first in the list,
        # since it may cause problems, like co-linear lines,
        # that need to be checked/fixed by later tests.
        self.test_list = [do_overlap_removal, do_cleanup]


def parse_glyph_list_arg(glyph_string):
    glyph_string = re.sub(r"[ \t\r\n,]+", ",", glyph_string)
    glyph_list = glyph_string.split(",")
    glyph_list = map(expand_names, glyph_list)
    glyph_list = filter(None, glyph_list)
    return glyph_list


class InlineHelpFormatter(argparse.RawDescriptionHelpFormatter):

    @staticmethod
    def __add_whitespace(indent_index, indent_space, arg):
        if indent_index == 0:
            return arg
        return (" " * indent_space) + arg

    def _split_lines(self, arg, width):
        arg_rows = arg.splitlines()
        for line_index, text_line in enumerate(arg_rows):
            search = re.search(r'\s*[0-9\-]{0,}\.?\s*', text_line)
            if text_line.strip() == "":
                arg_rows[line_index] = " "
            elif search:
                indent_space = search.end()
                lines = [self.__add_whitespace(i, indent_space, x) for i, x in
                         enumerate(textwrap.wrap(text_line, width))]
                arg_rows[line_index] = lines

        # The  [''] adds an empty line between arguments.
        return [item for sublist in arg_rows for item in sublist] + ['']


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=InlineHelpFormatter,
        description=__doc__
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '-q',
        '--quiet-mode',
        action='store_true',
        help='run in quiet mode'
    )
    parser.add_argument(
        '--no-overlap-checks',
        action='store_true',
        help='turn off path overlap checks'
    )
    parser.add_argument(
        '--no-basic-checks',
        action='store_true',
        help='turn off the following checks:\n'
             '- coincident points\n'
             '- colinear line segments\n'
             '- flat curves (straight segments with unnecessary control '
             'points)'
    )
    parser.add_argument(
        '--min-area',
        metavar='NUMBER',
        type=int,
        default=25,
        help='minimum area for a tiny outline\n'
             'Default is %(default)s square units. Subpaths with a bounding '
             'box less than this will be reported, and deleted if the -e '
             'option is used.'
    )
    parser.add_argument(
        '--tolerance',
        metavar='NUMBER',
        type=int,
        default=0,
        help='maximum allowed deviation from a straight line\n'
             'Default is %(default)s design space units. This is used to '
             'test whether the control points of a curve define a flat '
             'curve, and whether colinear line segments define a '
             'straight line.'
    )
    parser.add_argument(
        '-g',
        '--glyph-list',
        help='specify a list of glyphs to check\n'
             'Check only the specified list of glyphs. The list must be '
             'comma-delimited. The glyph IDs may be glyph indexes '
             'or glyph names. There must be no white-space in the '
             'glyph list.\n'
             'Example:\n'
             '    checkoutlinesufo -g A,B,C,69 MyFont.ufo'
    )
    parser.add_argument(
        '-f',
        '--glyph-file',
        metavar='FILE_PATH',
        type=validate_path,
        help='specify a file containing a list of glyphs to check\n'
             'Check only specific glyphs listed in a file. The '
             'file must contain a comma-delimited list of glyph '
             'identifiers. Any number of space, tab, and new-line '
             'characters are permitted between glyph names and commas.'
    )
    parser.add_argument(
        '-d',
        '--decimal',
        action='store_true',
        help='do NOT round point coordinates to integer'
    )
    parser.add_argument(
        '-e',
        '--error-correction-mode',
        action='store_true',
        help='correct reported problems\n'
             'The original font is modified. If the font is UFO, the outlines '
             'in the default layer are left untouched and a layer containing '
             'the modified version of the glyphs is added. The new layer is '
             f"named: '{PROCD_GLYPHS_LAYER}'"
    )
    parser.add_argument(
        '-o',
        '--output-file',
        metavar='FILE_PATH',
        nargs='?',
        help='Specify an output file to save to. RECOMMENDED for non-UFO '
             'inputs since the default behavior will overwrite the input.'
    )

    ufo_opts = parser.add_argument_group('UFO-only options')
    ufo_opts.add_argument(
        '-w',
        '--write-to-default-layer',
        action='store_true',
        help='write the modified glyphs to the default layer instead of '
             f"'{PROCD_GLYPHS_LAYER}'"
    )
    ufo_opts.add_argument(
        '--clear-hash-map',
        action='store_true',
        help='delete the hashes file\n'
             'By default, a file containing compact descriptions '
             '(a.k.a. hashes) of the glyphs is saved inside the '
             'UFO on the first time the font is checked. Storing '
             'these hashes allows checkoutlinesufo to be much faster '
             'on successive checks of the same UFO (because the '
             'tool will skip processing glyphs that were not '
             'modified since the last time the font was checked).'
    )
    ufo_opts.add_argument(
        '--all',
        action='store_true',
        help='force all glyphs to be processed\n'
             'Makes the tool ignore the stored hashes thus checking all the '
             'glyphs, even if they have already been processed.'
    )
    parser.add_argument(
        'font_path',
        metavar='FONT_PATH',
        type=validate_path,
        help='Path to UFO, OTF, CFF, or Type 1 font'
    )

    parsed_args = parser.parse_args(args)

    font_format = get_font_format(parsed_args.font_path)
    if font_format not in ('UFO', 'OTF', 'CFF', 'PFA', 'PFB', 'PFC'):
        parser.error('Font format is not supported.')

    if parsed_args.output_file:
        src = parsed_args.font_path
        dst = parsed_args.output_file
        if font_format == 'UFO':
            shutil.rmtree(dst, ignore_errors=True)
            shutil.copytree(src, dst)
        else:
            shutil.copy(src, dst)
        parsed_args.font_path = dst

    options = COOptions()

    if parsed_args.glyph_list:
        options.glyph_list += parse_glyph_list_arg(parsed_args.glyph_list)

    if parsed_args.glyph_file:
        gf = open(parsed_args.glyph_file)
        glyph_string = gf.read()
        glyph_string = glyph_string.strip()
        gf.close()
        options.glyph_list += parse_glyph_list_arg(glyph_string)

    if parsed_args.no_overlap_checks:
        options.test_list.remove(do_overlap_removal)
    if parsed_args.no_basic_checks:
        options.test_list.remove(do_cleanup)

    options.font_path = parsed_args.font_path
    options.font_format = font_format
    options.allow_changes = parsed_args.error_correction_mode
    options.quiet_mode = parsed_args.quiet_mode
    options.min_area = parsed_args.min_area
    options.tolerance = parsed_args.tolerance
    options.allow_decimal_coords = parsed_args.decimal
    options.check_all = parsed_args.all
    options.clear_hash_map = parsed_args.clear_hash_map
    options.write_to_default_layer = parsed_args.write_to_default_layer

    return options


def get_glyph_id(glyph_tag, font_glyph_list):
    glyph_id = None
    try:
        glyph_id = int(glyph_tag)
    except IndexError:
        pass
    except ValueError:
        try:
            glyph_id = font_glyph_list.index(glyph_tag)
        except IndexError:
            pass
        except ValueError:
            pass
    return glyph_id


def expand_names(glyph_name):
    glyph_range = glyph_name.split("-")
    if len(glyph_range) > 1:
        g1 = expand_names(glyph_range[0])
        g2 = expand_names(glyph_range[1])
        glyph_name = "%s-%s" % (g1, g2)

    elif glyph_name[0] == "/":
        glyph_name = "cid" + glyph_name[1:].zfill(5)
        if glyph_name == "cid00000":
            glyph_name = ".notdef"

    elif glyph_name.startswith("cid") and (len(glyph_name) < 8):
        glyph_name = "cid" + glyph_name[3:].zfill(5)
        if glyph_name == "cid00000":
            glyph_name = ".notdef"

    return glyph_name


def get_glyph_names(glyph_tag, font_glyph_list, font_file_name):
    glyph_name_list = []
    range_list = glyph_tag.split("-")
    prev_gid = get_glyph_id(range_list[0], font_glyph_list)
    if prev_gid is None:
        if len(range_list) > 1:
            print("\tWarning: glyph ID <%s> in range %s from glyph selection "
                  "list option is not in font. <%s>." %
                  (range_list[0], glyph_tag, font_file_name))
        else:
            print("\tWarning: glyph ID <%s> from glyph selection list option "
                  "is not in font. <%s>." % (range_list[0], font_file_name))
        return None
    try:
        glyph_name = font_glyph_list[prev_gid]
    except IndexError:
        print("\tWarning: glyph ID <%s> from glyph selection list option is "
              "not in font. <%s>." % (prev_gid, font_file_name))
        return None
    glyph_name_list.append(glyph_name)

    for glyph_tag_2 in range_list[1:]:
        gid = get_glyph_id(glyph_tag_2, font_glyph_list)
        if gid is None:
            print("\tWarning: glyph ID <%s> in range %s from glyph selection "
                  "list option is not in font. <%s>." %
                  (glyph_tag_2, glyph_tag, font_file_name))
            return None
        for i in range(prev_gid + 1, gid + 1):
            glyph_name_list.append(font_glyph_list[i])
        prev_gid = gid

    return glyph_name_list


def filter_glyph_list(options_glyph_list, font_glyph_list, font_file_name):
    # Return the list of glyphs which are in the intersection of the argument
    # list and the glyphs in the font.
    # Complain about glyphs in the argument list which are not in the font.
    glyph_list = []
    for glyph_tag in options_glyph_list:
        glyph_names = \
            get_glyph_names(glyph_tag, font_glyph_list, font_file_name)
        if glyph_names is not None:
            glyph_list.extend(glyph_names)
    return glyph_list


def get_digest(digest_glyph):
    """copied from robofab ObjectsBase.py.
    """
    mp = DigestPointPen()
    digest_glyph.drawPoints(mp)
    digest = list(mp.getDigestPointsOnly(needSort=False))
    digest.append((len(digest_glyph.contours), 0))
    return digest


def remove_coincident_points(bool_glyph, changed, msg):
    """ Remove coincident points.
    # a point is (segment_type, pt, smooth, name).
    # e.g. (u'curve', (138, -92), False, None)
    """
    for contour in bool_glyph.contours:
        op = contour._points[0]
        i = 1
        num_ops = len(contour._points)
        while i < num_ops:
            point = contour._points[i]
            segment_type = point[0]
            if segment_type is not None:
                last_op = op
                op = point
                if last_op[1] == op[1]:
                    # Coincident point. Remove it.
                    msg.append("Coincident points at (%s,%s)." % (op[1]))
                    changed = True
                    del contour._points[i]
                    num_ops -= 1
                    if segment_type == "curve":
                        # Need to delete the previous two points as well.
                        i -= 1
                        del contour._points[i]
                        i -= 1
                        del contour._points[i]
                        num_ops -= 2
                    # Note that we do not increment i - this will
                    # now index the next point.
                    continue
            i += 1
    return changed, msg


def remove_tiny_sub_paths(bool_glyph, min_area, msg):
    """
    Removes tiny subpaths that are created by overlap removal when the start
    and end path segments cross each other, rather than meet.
    """
    num_contours = len(bool_glyph.contours)
    ci = 0
    while ci < num_contours:
        contour = bool_glyph.contours[ci]
        ci += 1
        on_line_pts = filter(lambda pt: pt[0] is not None, contour._points)
        num_pts = len(list(on_line_pts))
        if num_pts <= 4:
            c_bounds = contour.bounds
            c_area = (c_bounds[2] - c_bounds[0]) * (c_bounds[3] - c_bounds[1])
            if c_area < min_area:
                ci -= 1
                num_contours -= 1
                msg.append(
                    "Contour %s is too small: bounding box is less than "
                    "minimum area. Start point: (%s)." %
                    (ci, contour._points[0][1]))
                del bool_glyph.contours[ci]
    return msg


def is_colinear_line(b3, b2, b1, tolerance=0):
    # b1-b3 are three points  [x, y] to test.
    # b1 = start point of first line,
    # b2 is start point of second,
    # b3 is end point of second line.
    # tolerance is how tight the match must be in design units, default 0.
    x1 = b1[-2]
    y1 = b1[-1]
    x2 = b2[-2]
    y2 = b2[-1]
    x3 = b3[-2]
    y3 = b3[-1]

    dx2 = x2 - x1
    dy2 = y2 - y1
    if (dx2 == 0) and (dy2 == 0):
        return True

    if (x3 - x1) == 0:
        if abs(dx2) > tolerance:
            return False
        else:
            return True

    if (y3 - y1) == 0:
        if abs(dy2) > tolerance:
            return False
        else:
            return True

    dx3 = x3 - x2
    dy3 = y3 - y2
    if (dx3 == 0) and (dy3 == 0):
        return False

    if dx2 == 0:
        if abs(x3 - x1) > tolerance:
            return False
        else:
            return True

    if dy2 == 0:
        if abs(y3 - y1) > tolerance:
            return False
        else:
            return True

    a = float(dy2) / dx2
    b = y1 - a * x1
    if abs(a) <= 1:
        y3t = a * x3 + b
        if abs(y3t - y3) > tolerance:
            return False
        else:
            return True
    else:
        x3t = float(y3 - b) / a
        if abs(x3t - x3) > tolerance:
            return False
        else:
            return True


def remove_flat_curves(new_glyph, changed, msg, options):
    """ Remove flat curves.
    # a point is (segment_type, pt, smooth, name).
    # e.g. (u'curve', (138, -92), False, None)
    """
    for contour in new_glyph.contours:
        num_ops = len(contour._points)
        i = 0
        while i < num_ops:
            p0 = contour._points[i]
            segment_type = p0[0]

            if segment_type == "curve":
                p3 = contour._points[i - 3]
                p2 = contour._points[i - 2]
                p1 = contour._points[i - 1]
                # fix flat curve.
                if is_colinear_line(
                        p3[1], p2[1], p0[1], options.tolerance):
                    if is_colinear_line(
                            p3[1], p1[1], p0[1], options.tolerance):
                        msg.append("Flat curve at (%s, %s)." % (p0[1]))
                        changed = True
                        p0 = list(p0)
                        p0[0] = "line"  # segment_type
                        contour._points[i] = tuple(p0)
                        num_ops -= 2
                        if i > 0:
                            del contour._points[i - 1]
                            del contour._points[i - 2]
                            i -= 2
                        else:
                            del contour._points[i - 1]
                            del contour._points[i - 1]
            i += 1
    return changed, msg


def remove_colinear_lines(new_glyph, changed, msg, options):
    """ Remove colinear line- curves.
    # a point is (segment_type, pt, smooth, name).
    # e.g. (u'curve', (138, -92), False, None)
    """
    for contour in new_glyph.contours:
        num_ops = len(contour._points)
        if num_ops < 3:
            # Need at least two line segments to have a co-linear line.
            continue
        i = 0
        while i < num_ops:
            p0 = contour._points[i]
            contour_type = p0[0]
            # Check for co-linear line.
            if contour_type == "line":
                p1 = contour._points[i - 1]
                # if p1 is a curve  point, no need to do colinear check!
                if p1[0] == "line":
                    p2 = contour._points[i - 2]
                    if is_colinear_line(
                            p2[1], p1[1], p0[1], options.tolerance):
                        msg.append("Colinear line at (%s, %s)." % (p1[1]))
                        changed = True
                        del contour._points[i - 1]
                        num_ops -= 1
                        if num_ops < 3:
                            # Need at least two line segments
                            # to have a co-linear line.
                            break
                        i -= 1
            i += 1
    return changed, msg


def split_touching_paths(new_glyph):
    """ This hack fixes a design difference between the Adobe checkoutlines
    logic and booleanGlyph, and is used only when comparing the two. With
    checkoutlines, if only a single point on a contour lines is coincident
    with the path of the another contour, the paths are NOT merged. With
    booleanGlyph, they are merged. An example is the vertex of a diamond
    shape having the same y coordinate as a horizontal line in another path,
    but no other overlap. However, this works only when a single point on one
    contour is coincident with another contour, with no other overlap. If
    there is more than one point of contact, then result is separate inner
    (unfilled) contour. This logic works only when the touching contour is
    merged with the other contour.
    """
    num_paths = len(new_glyph.contours)
    i = 0
    while i < num_paths:
        contour = new_glyph.contours[i]
        num_pts = len(contour._points)
        sort_list = [[contour._points[j], j] for j in range(num_pts)]
        sort_list = filter(lambda entry: entry[0][0] is not None, sort_list)
        sort_list.sort()
        j = 1
        len_sort_list = len(sort_list)
        have_new_contour = False
        while j < len_sort_list:
            coords1, j1 = sort_list[j][0][1], sort_list[j][1]
            coords0, j0 = sort_list[j - 1][0][1], sort_list[j - 1][1]
            if (num_pts != (j1 + 1)) and \
                    ((j1 - j0) > 1) and \
                    (coords0 == coords1):
                # If the contour has two non-contiguous points that are
                # coincident, and are not the start and end points, then we
                # have  contour that can be pinched off. I don't bother to
                # test that j0 is the start point, as UFO paths are always
                # closed.
                new_contour_pts = contour._points[j0:j1 + 1]
                np0 = new_contour_pts[0]
                if np0[0] == 'curve':
                    new_contour_pts[0] = ('line', np0[1], np0[2], np0[3])
                new_contour = booleanOperations.booleanGlyph.BooleanContour()
                new_contour._points = new_contour_pts
                new_contour._clockwise = new_contour._get_clockwise()
                # print("i", i)
                # print("coords1, j1", coords1, j1)
                # print("coords0, j0 ", coords0, j0)
                # print("new_contour._clockwise", repr(new_contour._clockwise))
                if 1:  # new_contour.clockwise == contour.clockwise:
                    # print("un-merging", new_contour._points,
                    #       repr(new_contour._clockwise))
                    del contour._points[j0 + 1:j1 + 1]
                    new_glyph.contours.insert(i + 1, new_contour)
                    have_new_contour = True
                    num_paths += 1
                    break
            j += 1
        if have_new_contour:
            continue
        i += 1


def round_point(pt):
    return int(pt[0]), int(pt[1])


def do_overlap_removal(bool_glyph, changed, msg, options):
    changed, msg = remove_coincident_points(bool_glyph, changed, msg)
    options.remove_coincident_points_done = True
    changed, msg = remove_flat_curves(bool_glyph, changed, msg, options)
    changed, msg = remove_colinear_lines(bool_glyph, changed, msg, options)
    options.remove_flat_curves_done = True
    # I need to fix these in the source, or the old vs new digests will differ,
    # as BooleanOperations removes these even if it does not do overlap
    # removal.
    old_digest = sorted(get_digest(bool_glyph))
    old_digest = [round_point(pt) for pt in old_digest]
    new_digest = []
    prev_digest = old_digest
    new_glyph = bool_glyph

    while new_digest != prev_digest:
        # This is a hack to get around a bug in booleanGlyph. Consider an M
        # sitting on a separate rectangular crossbar contour, the bottom of
        # the M legs being co-linear with the top of the cross bar. pyClipper
        # will merge only one of the co-linear lines with each call to
        # removeOverlap(). I suspect that this bug is in pyClipper, but
        # haven't yet looked.
        prev_digest = new_digest
        new_glyph = new_glyph.removeOverlap()
        new_digest = sorted(get_digest(new_glyph))
        # The new path points sometimes come back with very small
        # fractional parts to to rounding issues.
        new_digest = [round_point(pt) for pt in new_digest]

    # Can't use change in path number to see if something has changed
    # - overlap removal can add and subtract paths.
    if str(old_digest) != str(new_digest):
        changed = True
        msg.append("There is an overlap.")
    # Don't need to remove coincident points again.
    # These are not added by overlap removal.
    # Tiny subpaths are added by overlap removal.
    msg = remove_tiny_sub_paths(new_glyph, options.min_area, msg)
    return new_glyph, changed, msg


def do_cleanup(new_glyph, changed, msg, options):

    if new_glyph is None:
        return new_glyph, changed, msg

    # Note that these remove_coincident_points_done and remove_flat_curves_done
    # get called only if do_overlap_removal is NOT called.
    if not options.remove_coincident_points_done:
        changed, msg = remove_coincident_points(new_glyph, changed, msg)
        options.remove_coincident_points_done = True
    if not options.remove_flat_curves_done:
        changed, msg = remove_flat_curves(new_glyph, changed, msg, options)
    # I call remove_colinear_lines even when do_overlap_removal is called,
    # as the latter can introduce new co-linear points.
    changed, msg = remove_colinear_lines(new_glyph, changed, msg, options)

    return new_glyph, changed, msg


def set_max_p(contour):
    # Find the topmost point
    # If more than one point has the same Y, choose the leftmost.
    # Used to sort contours.
    max_p = contour[0]
    for point in contour:
        if point.segmentType is None:
            continue
        if max_p.y < point.y:
            max_p = point
        elif max_p.y == point.y:
            if max_p.x < point.y:
                max_p = point
    contour.maxP = max_p


def sort_contours(c1, c2):
    if not hasattr(c1, 'maxP'):
        set_max_p(c1)
    if not hasattr(c2, 'maxP'):
        set_max_p(c2)

    if c1.maxP.y > c2.maxP.y:
        return 1
    elif c1.maxP.y < c2.maxP.y:
        return -1

    if c1.maxP.x > c2.maxP.x:
        return 1
    elif c1.maxP.x < c2.maxP.x:
        return -1

    lc1 = len(c1)
    lc2 = len(c2)
    if lc1 > lc2:
        return 1
    elif lc1 < lc2:
        return -1
    return 0


def restore_contour_order(fixed_glyph, original_contours):
    """ The pyClipper library first sorts all the outlines by x position,
    then y position. I try to undo that, so that un-touched contours will end
    up in the same order as the in the original, and any conbined contours
    will end up in a similar order. The reason I try to match new contours
    to the old is to reduce arbitraryness in the new contour order between
    similar fonts. I can't completely avoid this, but I can reduce how often
    it happens.
    """
    if len(fixed_glyph) == 0:
        return

    new_contours = list(fixed_glyph)
    if len(new_contours) > 1:
        new_contours.sort(key=cmp_to_key(sort_contours))
    else:
        set_max_p(new_contours[0])
    new_index_list = range(len(new_contours))
    new_list = [[i, new_contours[i]] for i in new_index_list]
    old_index_list = range(len(original_contours))
    old_list = [[i, original_contours[i]] for i in old_index_list]
    order_list = []

    # Match contours that have not changed.
    # This will fix the order of the contours that have not been touched.
    num_contours = len(new_list)
    if num_contours > 0:  # If the new contours aren't already all matched..
        for i in range(num_contours):
            ci, contour = new_list[i]
            for j in old_index_list:
                ci2, old_contour = old_list[j]
                # skip if is not the same set of values.
                if len(old_contour) != len(contour):
                    continue
                if old_contour[0] != old_contour[0]:
                    continue
                order_list.append([ci2, ci])
                new_list[i] = None
                break

    new_list = [item for item in new_list if item]
    num_contours = len(new_list)
    # Check each extreme for a match.
    if num_contours > 0:
        # ctr_starts tracks start points per contour
        ctr_starts = {n: [] for n in range(len(fixed_glyph))}
        for i in range(num_contours):
            ci, contour = new_list[i]
            max_p = contour.maxP
            # Now search the old contour list.
            for j in old_index_list:
                ci2, old_contour = old_list[j]
                matched = False
                for old_point in old_contour:
                    if old_point.segmentType is None:
                        continue
                    if (max_p.x == old_point.x) and (max_p.y == old_point.y):
                        new_list[i] = None
                        order_list.append([ci2, ci])
                        matched = True
                        # See if we can set the start point in the new contour
                        # to match the old one.
                        if not ((old_contour[0].x == contour[0].x) and
                                (old_contour[0].y == contour[0].y)):
                            old_start_point = old_contour[0]
                            for pi, point in enumerate(contour):
                                if (point.x == old_start_point.x) \
                                        and (point.y == old_start_point.y) \
                                        and point.segmentType is not None:
                                    try:
                                        msg = ("Failed assertion: duplicated "
                                               f"start point on contour {ci} "
                                               f"at {point.x}, {point.y} of "
                                               f"glyph {fixed_glyph.name}.")
                                        assert not ctr_starts[ci], msg
                                    except KeyError:
                                        msg = (f"Contour index {ci} out "
                                               "of range on updated glyph "
                                               f"{fixed_glyph.name}")
                                        raise KeyError(msg)
                                    contour.setStartPoint(pi)
                                    ctr_starts[ci].append(pi)

                        break
                if matched:
                    break
        new_list = [item for item in new_list if item]
        num_contours = len(new_list)

    # If the algorithm didn't work for some contours,
    # just add them on the end.
    if num_contours != 0:
        ci2 = len(new_contours)
        for ci, _contour in new_list:
            order_list.append([ci2, ci])

    # Now re-order the new list
    order_list.sort()
    new_contour_list = []
    for _ci2, ci in order_list:
        new_contour_list.append(new_contours[ci])

    fixed_glyph.clearContours()
    for contour in new_contour_list:
        fixed_glyph.appendContour(contour)


RE_SPACE_PATTERN = re.compile(
    r"space|uni(00A0|1680|180E|202F|205F|3000|FEFF|200[0-9AB])")


def run(args=None):
    options = get_options(args)
    font_path = options.font_path
    font_format = options.font_format
    font_file = FontFile(font_path, font_format)
    use_hash_map = True if font_format == 'UFO' else False
    defcon_font = font_file.open(use_hash_map)

    # We allow use of a hash map to skip glyphs only if fixing glyphs
    if options.clear_hash_map:
        font_file.clear_hash_map()
        return

    if defcon_font is None:
        print("Could not open file: %s." % font_path)
        return

    if not options.glyph_list:
        glyph_list = list(defcon_font.keys())
    else:
        if not defcon_font.glyphOrder:
            raise FocusFontError(
                "Error: public.glyphOrder is empty or missing "
                "from lib.plist file of %s" % font_path)
        else:
            glyph_list = filter_glyph_list(
                options.glyph_list, defcon_font.glyphOrder, font_path)
    if not glyph_list:
        raise FocusFontError(
            "Error: selected glyph list is empty for font <%s>." % font_path)

    if (font_format == 'UFO') and (not options.write_to_default_layer):
        try:
            processed_layer = defcon_font.layers[PROCD_GLYPHS_LAYER_NAME]
        except KeyError:
            processed_layer = defcon_font.newLayer(PROCD_GLYPHS_LAYER_NAME)
    else:
        processed_layer = None
        font_file.save_to_default_layer = True

    font_changed = False
    seen_glyph_count = 0
    processed_glyph_count = 0

    max_length = max([len(g) for g in glyph_list])
    fmt = '{:<%d}' % max_length
    with trange(len(glyph_list)) as t:
        for i in t:
            glyph_name = sorted(glyph_list)[i]
            t.set_description('Checking outlines for %s' %
                              fmt.format(glyph_name))
            changed = False
            seen_glyph_count += 1
            msg = []

            if glyph_name not in defcon_font:
                continue

            # font_file.check_skip_glyph updates the hash map for the glyph,
            # so we call it even when the  '-all' option is used.
            skip = font_file.check_skip_glyph(glyph_name, options.check_all)
            # Note: this will delete glyphs from the processed layer,
            #       if the glyph hash has changed.
            if skip:
                continue
            processed_glyph_count += 1

            defcon_glyph = defcon_font[glyph_name]
            if defcon_glyph.components:
                defcon_glyph.decomposeAllComponents()
            new_glyph = booleanOperations.booleanGlyph.BooleanGlyph(
                defcon_glyph)
            if len(new_glyph) == 0:
                # Complain about empty glyph only if it is not a space glyph.
                if not RE_SPACE_PATTERN.search(glyph_name):
                    msg = ["has no contours"]
                else:
                    msg = []
            else:
                for test in options.test_list:
                    if test is not None:
                        new_glyph, changed, msg = \
                            test(new_glyph, changed, msg, options)

            if not options.quiet_mode:
                if msg:
                    tqdm.write('%s %s' % (glyph_name, ' '.join(msg)))
            if changed and options.allow_changes:
                font_changed = True
                original_contours = list(defcon_glyph)
                if font_file.save_to_default_layer:
                    fixed_glyph = defcon_glyph
                    fixed_glyph.clearContours()
                else:
                    # this will replace any pre-existing glyph:
                    processed_layer.newGlyph(glyph_name)
                    fixed_glyph = processed_layer[glyph_name]
                    fixed_glyph.width = defcon_glyph.width
                    fixed_glyph.height = defcon_glyph.height
                    fixed_glyph.unicodes = defcon_glyph.unicodes
                point_pen = fixed_glyph.getPointPen()
                new_glyph.drawPoints(point_pen)
                if options.allow_decimal_coords:
                    for contour in fixed_glyph:
                        for point in contour:
                            point.x = round(point.x, 3)
                            point.y = round(point.y, 3)
                else:
                    for contour in fixed_glyph:
                        for point in contour:
                            point.x = int(round(point.x))
                            point.y = int(round(point.y))

                # JH May 2020: remove overlap can leave some coincident points
                # we use thresholdAttrGlyph (modified thresholdPen) to remove
                # them prior to restore_contour_order.
                thresholdAttrGlyph(fixed_glyph, 1)

                restore_contour_order(fixed_glyph, original_contours)

            # The following is needed when the script is called from another
            # script with Popen():
            sys.stdout.flush()
            if i == len(glyph_list) - 1:
                t.set_description("Finished checkoutlinesufo")

    # update layer plist: the hash check call may have deleted processed layer
    # glyphs because the default layer glyph is newer.

    # At this point, we may have deleted glyphs in the
    # processed layer.writer.getGlyphSet()
    # will fail unless we update the contents.plist file to match.
    if options.allow_changes:
        ufotools.validateLayers(font_path, False)

    if not font_changed:
        # Even if the program didn't change any glyphs,
        # we should still save updates to the src glyph hash file.
        font_file.close()
    else:
        font_file.save()

    if processed_glyph_count != seen_glyph_count:
        print("Skipped %s of %s glyphs." %
              (seen_glyph_count - processed_glyph_count, seen_glyph_count))
    if not options.quiet_mode:
        print("Done with font")


def main():
    try:
        run(sys.argv[1:])
    except (FocusOptionParseError, FocusFontError) as focus_error:
        print("Quitting after error.", focus_error)


if __name__ == '__main__':
    main()



import pytest
import re
import subprocess
import time

from runner import main as runner
from differ import main as differ, SPLIT_MARKER
from test_utils import get_expected_path

TOOL = 'spot'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

@pytest.mark.parametrize('arg', ['-h', '-u'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-z', '-foo'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 1


@pytest.mark.parametrize('args, exp_filename', [
    (['T'], 'list_tables.txt'),
    (['F'], 'list_features.txt'),
    (['G'], 'glyph_proof.ps'),  # same as dump_CFF_=7
    (['G', 'd'], 'glyph_proof_no_header.ps'),
    (['G', 'd', 'br'], 'glyph_proof_no_header_multipage.ps'),
    (['Proof'], 'proof_all_features.ps'),
    (['t', '_BASE=4'], 'dump_BASE=4.txt'),
    (['t', '_BASE=5'], 'dump_BASE=5.txt'),
    (['t', '_GDEF=7'], 'dump_GDEF=7.txt'),
    (['t', '_GPOS=5'], 'dump_GPOS=5.txt'),
    (['t', '_GPOS=6'], 'dump_GPOS=6.txt'),
    (['t', '_GPOS=7'], 'dump_GPOS=7.txt'),
    (['t', '_GPOS=8'], 'dump_GPOS=8.ps'),
    (['t', '_GSUB=4'], 'dump_GSUB=4.txt'),
    (['t', '_GSUB=5'], 'dump_GSUB=5.txt'),
    (['t', '_GSUB=7'], 'dump_GSUB=7.txt'),
    (['t', '_GSUB=8'], 'dump_GSUB=8.ps'),
    (['t', '_cmap=5'], 'dump_cmap=5.txt'),
    (['t', '_cmap=6'], 'dump_cmap=6.txt'),
    (['t', '_cmap=7'], 'dump_cmap=7.txt'),
    (['t', '_cmap=8'], 'dump_cmap=8.txt'),
    (['t', '_cmap=9'], 'dump_cmap=9.ps'),
    (['t', '_cmap=10'], 'dump_cmap=10.ps'),
    (['t', '_cmap=11'], 'dump_cmap=11.txt'),
    (['t', '_hmtx=5'], 'dump_hmtx=5.txt'),
    (['t', '_hmtx=6'], 'dump_hmtx=6.txt'),
    (['t', '_hmtx=7'], 'dump_hmtx=7.txt'),
    (['t', '_hmtx=8'], 'dump_hmtx=8.txt'),
    (['t', '_name=2'], 'dump_name=2.txt'),
    (['t', '_name=3'], 'dump_name=3.txt'),
    (['t', '_name=4'], 'dump_name=4.txt'),
    (['t', '_CFF_=6'], 'dump_CFF_=6.ps'),
    (['t', '_CFF_=6', 'a'], 'dump_CFF_=6a.ps'),
    (['t', '_CFF_=6', 'c'], 'dump_CFF_=6c.ps'),
    (['t', '_CFF_=6', 'R'], 'dump_CFF_=6R.ps'),
    (['t', '_CFF_=6', 'g2,5-7'], 'dump_CFF_=6g.ps'),
    (['t', '_CFF_=6', 's1.5,0.5'], 'dump_CFF_=6s.ps'),
    (['t', '_CFF_=6', 'b-10,-10,200,500', 'g4'], 'dump_CFF_=6b.ps'),
    (['t', '_CFF_=7'], 'dump_CFF_=7.ps'),
    (['t', '_CFF_=7', 'g2,5-7'], 'dump_CFF_=7g.ps'),
    (['t', '_CFF_=8'], 'dump_CFF_=8.ps'),
    (['t', '_CFF_=8', 'g2,5-7'], 'dump_CFF_=8g.ps'),
])
def test_options(args, exp_filename):
    if 'CFF_=7' in exp_filename or exp_filename == 'glyph_proof.ps':
        skip = ['0 743.4 moveto (CFF:']
    elif 'CFF_=6' in exp_filename:
        skip = ['318 764 moveto (' + SPLIT_MARKER + '72 750 moveto (']
    else:
        skip = []
    if skip:
        skip.insert(0, '-s')
    actual_path = runner(CMD + ['-s', '-f', 'black.otf', '-o'] + args)
    expected_path = get_expected_path(exp_filename)
    assert differ([expected_path, actual_path] + skip)


@pytest.mark.parametrize('table', ['GSUB', 'GPOS'])
def test_feature_tags_bug691(table):
    exp_filename = f'SourceCodePro-Regular_{table}.ps'
    actual_path = runner(CMD + ['-s', '-f', 'SourceCodePro-Regular.otf',
                                '-o', 't', f'_{table}=8'])
    expected_path = get_expected_path(exp_filename)
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('font_format', ['otf', 'ttf'])
def test_long_glyph_name_bug373(font_format):
    file_name = 'long_glyph_name.' + font_format
    actual_path = runner(CMD + ['-s', '-o', 't', '_GSUB=7', '-f', file_name])
    expected_path = get_expected_path(f'bug373_{font_format}.txt')
    assert differ([expected_path, actual_path])


def test_buffer_overrun_bug465():
    """ Fix bug where a fixed length string buffer was overrun. The test
    font was built by building an otf from makeotf_data/input/t1pfa.pfa,
    dumping this to ttx, and then hand-editing the ttx file to have
    empty charstrings, and a ContextPos lookup type 7, format 2.
    Hand-editing the ttx was needed as makeotf does not build this
    lookup type, nor this post rule format. This test validates that the
    current spot code correctly reports the test font GPOS table."""
    file_name = "bug465/bug465.otf"
    expected_path = get_expected_path('bug465_otf.txt')
    actual_path = runner(
        CMD + ['-s', expected_path, '-o', 't', '_GPOS=7', '-f', file_name])
    assert differ([expected_path, actual_path])


def test_full_unicode_ranges_bug813():
    font_name = 'AdobeBlack2VF.otf'
    actual_path = runner(CMD + ['-s', '-o', 't', '_OS/2', '-f', font_name])
    expected_path = get_expected_path('AdobeBlack2VF_OS2.txt')
    assert differ([expected_path, actual_path])


@pytest.mark.parametrize('args, input_filename', [
    (['t', '_glyf=6'], 'black.ttf'),
    (['t', '_glyf=7'], 'black.ttf'),
    (['t', '_CFF_=7'], 'black.otf'),
])
def test_date_and_time(args, input_filename):
    """
    There are three different places in the spot code where time() and
    localtime_r() are called. This test is meant to exercise all three of them
    and to confirm that the time is correctly set.
    """
    actual_path = runner(CMD + ['-s', '-f', input_filename, '-o'] + args)
    found_date = False
    with open(actual_path) as output_file:
        lines = output_file.readlines()
        now = time.time()
        for line in lines:
            line = line.strip()
            if line.startswith('318 764 moveto'):
                # this is the pattern for glyf=6
                file_timestamp = re.split(r'[()]', line)[1]
                file_time = time.mktime(time.strptime(file_timestamp))
                found_date = True
            elif line.startswith('0 743.4 moveto ('):
                # this is the pattern for glyf=7 and CFF_=7
                file_timestamp = ' '.join(re.split(r'[\s()]+', line)[-3:-1])
                file_time = time.mktime(
                    time.strptime(file_timestamp, '%m/%d/%y %H:%M'))
                found_date = True
            if found_date:
                # The difference in time between when the file was made
                # and now should be less than an hour. That should be enough
                # even for some of the slower CI processes. ;-)
                hours_diff = abs(now - file_time) / 3600
                assert(hours_diff < 1)
                break

    # Make sure we actually found a date and checked it.
    # (This is to catch the case of an empty or incomplete output file.)
    assert found_date



# Copyright 2017 Adobe. All rights reserved.
"""
Builds master source OpenType/CFF fonts from a designspace file
and UFO master source fonts.
"""

# The script makes the following assumptions:
# 1. The AFDKO is installed; the script calls 'tx', 'sfntedit' and 'makeotf').
# 2. Each master source font is in a separate directory from the other master
#    fonts, and each of these directories provides the default metadata files
#    expected by the 'makeotf' command:
#    - features/features.fea
#    - font.ufo
#    - FontMenuNameDB within the directory or no more than 4 parent
#      directories up.
#    - GlyphAliasAndOrderDB within the directory or no more than 4 parent
#      directories up.
# The script will build a master source OpenType/CFF font for each master
# source UFO font, using the same file name, but with the extension '.otf'.

import argparse
import logging
import os
import sys

from fontTools.designspaceLib import (
    DesignSpaceDocument,
    DesignSpaceDocumentError,
)
from afdko.fdkutils import (
    run_shell_command,
    get_temp_file_path,
    validate_path,
)
from afdko.makeotf import main as makeotf

__version__ = '1.10.0'

logger = logging.getLogger(__name__)


class ShellCommandError(Exception):
    pass


def generalizeCFF(otfPath):
    tempFilePath = get_temp_file_path()

    if not run_shell_command(['tx', '-cff', '+b', '-std', '-no_opt',
                              otfPath, tempFilePath]):
        raise ShellCommandError

    if not run_shell_command(['sfntedit', '-a',
                              f'CFF ={tempFilePath}', otfPath]):
        raise ShellCommandError


def _split_makeotf_options(comma_str):
    """Converts a comma-separated string into a list.
       Adds back leading hyphen."""
    if not comma_str.startswith('-'):
        comma_str = '-' + comma_str
    return comma_str.split(',')


MKOT_OPT = '--mkot'


def _preparse_args(args):
    """Removes the leading hyphen from the '--mkot' option argument"""
    if args is None:
        args = sys.argv[1:]
    if MKOT_OPT not in args:
        return args
    opt_i = args.index(MKOT_OPT)
    try:
        arg_i = opt_i + 1
        mkot_arg = args[arg_i]
        if mkot_arg.startswith('-'):
            args[arg_i] = mkot_arg[1:]
    except IndexError:
        pass
    return args


def build_masters(opts):
    """
    Build master OTFs using supplied options.
    """
    logger.info("Reading designspace file...")
    ds = DesignSpaceDocument.fromfile(opts.dsPath)
    validateDesignspaceDoc(ds)
    master_paths = [s.path for s in ds.sources]

    logger.info("Building local OTFs for master font paths...")
    dsDir = os.path.dirname(opts.dsPath)

    for master_path in master_paths:
        master_path = os.path.join(dsDir, master_path)
        otf_path = f"{os.path.splitext(master_path)[0]}.otf"

        result = makeotf(['-nshw', '-f', master_path, '-o', otf_path,
                          '-r', '-nS'] + opts.mkot)

        if result:
            raise Exception(f'makeotf return value: {result}')

        logger.info(f"Built OTF font for {master_path}")
        generalizeCFF(otf_path)


def get_options(args):
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__
    )
    parser.add_argument(
        '--version',
        action='version',
        version=__version__
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='verbose mode\n'
             'Use -vv for debug mode'
    )
    parser.add_argument(
        MKOT_OPT,
        metavar='OPTIONS',
        help="comma-separated set of 'makeotf' options",
        type=_split_makeotf_options,
        default=[]
    )
    parser.add_argument(
        '-d',
        '--designspace',
        metavar='PATH',
        dest='dsPath',
        type=validate_path,
        help='path to design space file',
        required=True
    )
    options = parser.parse_args(args)

    if not options.verbose:
        level = "WARNING"
    elif options.verbose == 1:
        level = "INFO"
    else:
        level = "DEBUG"
    logging.basicConfig(level=level)

    return options


def validateDesignspaceDoc(dsDoc, **kwArgs):
    """
    Validate the dsDoc DesignSpaceDocument object. Raises Exceptions if
    certain criteria are not met. These are above and beyond the basic
    validations in fontTools.designspaceLib and are specific to
    buildmasterotfs.
    """
    if dsDoc.sources:
        for src in dsDoc.sources:
            if not os.path.exists(src.path):
                raise DesignSpaceDocumentError(
                    f"Source file {src.path} does not exist")
    else:
        raise DesignSpaceDocumentError("Designspace file contains no sources.")


def main(args=None):
    args = _preparse_args(args)
    opts = get_options(args)

    try:
        build_masters(opts)
    except Exception as exc:
        logger.error(exc)
        return 1


if __name__ == "__main__":
    sys.exit(main())



import glob
import os
import pytest
import subprocess

from runner import main as runner
from differ import main as differ, SPLIT_MARKER
from test_utils import (get_input_path, get_expected_path, get_temp_file_path,
                        generate_ttx_dump, generate_spot_dumptables)

TOOL = 'makeotfexe'
CMD = ['-t', TOOL]


# -----
# Tests
# -----

def test_exit_no_option():
    # It's valid to run 'makeotfexe' without using any options,
    # but if a default-named font file ('font.ps') is NOT found
    # in the current directory, the tool exits with an error
    with pytest.raises(subprocess.CalledProcessError) as err:
        subprocess.check_call([TOOL])
    assert err.value.returncode == 1


@pytest.mark.parametrize('arg', ['-h', '-u'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-j', '--bogus'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 1


@pytest.mark.parametrize('args, ttx_fname', [
    ([], 'font_dev'),
    (['r'], 'font_rel'),
])
def test_build_font_and_check_messages(args, ttx_fname):
    actual_path = get_temp_file_path()
    expected_msg_path = get_expected_path(f'{ttx_fname}_output.txt')
    ttx_filename = f'{ttx_fname}.ttx'
    stderr_path = runner(CMD + [
        '-s', '-e', '-o', 'f', f'_{get_input_path("font.pfa")}',
                          'o', f'_{actual_path}'] + args)
    actual_ttx = generate_ttx_dump(actual_path)
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value=',
                   '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])
    assert differ([expected_msg_path, stderr_path])


@pytest.mark.parametrize('caret_format', [
    'bypos', 'byindex', 'mixed', 'mixed2', 'double', 'double2'])
def test_GDEF_LigatureCaret_bug155(caret_format):
    input_filename = 'bug155/font.pfa'
    feat_filename = f'bug155/caret-{caret_format}.fea'
    ttx_filename = f'bug155/caret-{caret_format}.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GDEF'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-l', '2'])


def test_useMarkFilteringSet_flag_bug196():
    input_filename = "bug196/font.pfa"
    feat_filename = "bug196/feat.fea"
    actual_path = get_temp_file_path()
    ttx_filename = "bug196.ttx"
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GSUB'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_useException_bug321():
    input_filename = "bug321/font.pfa"
    feat_filename = "bug321/feat.fea"
    actual_path = get_temp_file_path()
    ttx_filename = "bug321.ttx"
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GSUB', 'GPOS'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_mark_refer_diff_classes_bug416():
    input_filename = "bug416/font.pfa"
    feat_filename = "bug416/feat.fea"
    actual_path = get_temp_file_path()
    ttx_filename = "bug416.ttx"
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GPOS'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_DFLT_script_with_any_lang_bug438():
    """ The feature file bug438/feat.fea contains languagesystem
    statements for a language other than 'dflt' with the 'DFLT' script
    tag. With the fix, makeotfexe will build an OTF which is identical to
    'bug438.ttx'. Without the fix, it will fail to build an OTF."""
    input_filename = 'bug438/font.pfa'
    feat_filename = 'bug438/feat.fea'
    ttx_filename = 'bug438.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path)
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value=',
                   '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])


def test_glyph_not_in_font_bug492():
    input_filename = 'bug492/font.pfa'
    feat_filename = 'bug492/features.fea'
    otf_path = get_temp_file_path()

    stderr_path = runner(
        CMD + ['-s', '-e', '-o',
               'f', f'_{get_input_path(input_filename)}',
               'ff', f'_{get_input_path(feat_filename)}',
               'o', f'_{otf_path}'])

    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert (b'[ERROR] <SourceSans-Test> Glyph "glyph_not_found" not in font. '
            b'[') in output
    assert (b'syntax error at "a" [') not in output


def test_version_warning_bug610():
    input_filename = 'bug610/font.pfa'
    feat_filename = 'bug610/v0005.fea'
    otf_path = get_temp_file_path()

    stderr_path = runner(
        CMD + ['-s', '-e', '-o',
               'f', f'_{get_input_path(input_filename)}',
               'ff', f'_{get_input_path(feat_filename)}',
               'o', f'_{otf_path}'])

    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert (b"[WARNING] <SourceSans-Test> Major version number not in "
            b"range 1 .. 255") not in output


@pytest.mark.parametrize('feat_name, error_msg', [
    ('test_named_lookup',
        b"[FATAL] <SourceSans-Test> In feature 'last' positioning rules "
        b"cause an offset overflow (0x10020) to a lookup subtable"),
    ('test_singlepos_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'sps1' lookup 'lkup40' "
        b"positioning rules cause an offset overflow (0x10188) to a "
        b"lookup subtable"),
    ('test_class_pair_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'last' positioning rules "
        b"cause an offset overflow (0x10020) to a lookup subtable"),
    ('test_class_pair_class_def_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'krn1' lookup 'l1' pair "
        b"positioning rules cause an offset overflow (0x1001a) to a "
        b"class 2 definition table"),
    ('test_contextual_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'krn0' lookup 'lkup40' "
        b"chain contextual positioning rules cause an offset overflow "
        b"(0x10002) to a lookup subtable"),
    ('test_cursive_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'curs' lookup 'lk20' "
        b"cursive positioning rules cause an offset overflow "
        b"(0x1006e) to a cursive attach table"),
    ('test_mark_to_base_coverage_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'mrk1' mark to base "
        b"positioning rules cause an offset overflow (0x1002c) to a "
        b"base coverage table"),
    ('test_mark_to_base_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'mrk1' mark to base "
        b"positioning rules cause an offset overflow (0x10230) to a "
        b"lookup subtable"),
    ('test_mark_to_ligature_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'lig1' lookup 'lk0' mark "
        b"to ligature positioning rules cause an offset overflow (0x1053e) "
        b"to a lookup subtable"),
    ('test_singlesub1_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'tss2' lookup 'lkup258' "
        b"substitution rules cause an offset overflow (0x10002) to a "
        b"lookup subtable"),
    ('test_singlesub2_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'tss1' lookup 'lkup237' "
        b"substitution rules cause an offset overflow (0x10098) to a "
        b"lookup subtable"),
    ('test_multiplesub_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'mts1' lookup 'lkup238' "
        b"substitution rules cause an offset overflow (0x10056) to a "
        b"lookup subtable"),
    ('test_alternatesub_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'ats1' lookup 'lkup238' "
        b"substitution rules cause an offset overflow (0x1009c) to a "
        b"lookup subtable"),
    ('test_ligaturesub_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'lts1' lookup 'lkup238' "
        b"substitution rules cause an offset overflow (0x10016) to a "
        b"lookup subtable"),
    ('test_chaincontextualsub_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'cts1' substitution rules "
        b"cause an offset overflow (0x100ac) to a lookup subtable"),
    ('test_reversechaincontextualsub_subtable_overflow',
        b"[FATAL] <SourceSans-Test> In feature 'rts1' lookup 'lkup238' "
        b"reverse chain contextual substitution rules cause an offset "
        b"overflow (0x100a0) to a lookup subtable"),
    ('test_duplicate_single_sub',
        b"[NOTE] <SourceSans-Test> Removing duplicate single substitution "
        b"in standalone lookup 'test2': glyph1, glyph1"),
    ('test_conflicting_single_sub',
        b"[FATAL] <SourceSans-Test> Duplicate target glyph for single "
        b"substitution in standalone lookup 'test2': glyph1"),
    ('test_duplicate_alternate_sub',
        b"[FATAL] <SourceSans-Test> Duplicate target glyph for alternate "
        b"substitution in standalone lookup 'test2': glyph1"),
    ('test_duplicate_multiple_sub',
        b"[FATAL] <SourceSans-Test> Duplicate target glyph for multiple "
        b"substitution in standalone lookup 'test2': glyph1"),
    ('test_duplicate_ligature_sub',
        b"[NOTE] <SourceSans-Test> Removing duplicate ligature substitution "
        b"in standalone lookup 'test2'"),
    ('test_conflicting_ligature_sub',
        b"[FATAL] <SourceSans-Test> Duplicate target sequence but different "
        b"replacement glyphs in ligature substitutions in standalone "
        b"lookup 'test2'"),
    ('test_duplicate_single_pos',
        b"[NOTE] <SourceSans-Test> Removing duplicate single positioning "
        b"in standalone lookup 'test2'"),
    ('test_conflicting_single_pos',
        b"[FATAL] <SourceSans-Test> Duplicate single positioning glyph "
        b"with different values in standalone lookup 'test2'"),
    ('test_features_name_missing_win_dflt_sub',
        b"[FATAL] <SourceSans-Test> Missing Windows default name for "
        b"'featureNames' nameid 256 in feature 'ss01'"),
    ('test_cv_params_not_in_cvxx_sub',
        b"[FATAL] <SourceSans-Test> A 'cvParameters' block is only allowed "
        b"in Character Variant (cvXX) features; it is being used in "
        b"feature 'tst1'"),
    ('test_cv_params_missing_win_dflt_sub',
        b"[FATAL] <SourceSans-Test> Missing Windows default name for "
        b"'cvParameters' nameid 256 in feature 'cv01'"),
    ('test_size_withfamilyID_0_pos',
        b"[FATAL] <SourceSans-Test> 'size' feature must have 4 parameters if "
        b"sub family ID code is non-zero!"),
    ('test_size_withfamilyID_3_pos',
        b"[FATAL] <SourceSans-Test> 'size' feature must have 4 or 2 "
        b"parameters if sub family code is zero!"),
    ('test_sizemenuname_missing_win_dflt_pos',
        b"[FATAL] <SourceSans-Test> Missing Windows default name for "
        b"'sizemenuname' nameid 256 in 'size' feature"),
    ('test_kernpair_warnings_pos',
        b"[WARNING] <SourceSans-Test> Single kern pair occurring after class "
        b"kern pair in feature 'tst1'"),
    ('test_kernpair_warnings_pos',
        b"[WARNING] <SourceSans-Test> Start of new pair positioning subtable "
        b"forced by overlapping glyph classes in feature 'tst1'; some pairs "
        b"may never be accessed"),
    ('test_kernpair_warnings_pos',
        b"[NOTE] <SourceSans-Test> Removing duplicate pair positioning in "
        b"feature 'tst1': glyph3 glyph3"),
    ('test_kernpair_warnings_pos',
        b"[WARNING] <SourceSans-Test> Pair positioning has conflicting "
        b"statements in feature 'tst1'"),
    ('test_base_glyph_conflict_pos',
        b"[ERROR] <SourceSans-Test> MarkToBase or MarkToMark error in "
        b"feature 'tst1'. Another statement has already defined the anchors "
        b"and marks on glyph 'glyph5'."),
    ('test_base_glyph_conflict_pos',
        b"[WARNING] <SourceSans-Test> MarkToBase or MarkToMark error in "
        b"feature 'tst1'. Glyph 'glyph5' does not have an anchor point "
        b"for a mark class that was used in a previous statement in the "
        b"same lookup table. Setting the anchor point offset to 0."),
    ('test_mark_ligature_conflict_pos',
        b"[ERROR] <SourceSans-Test> MarkToLigature error in feature 'tst1'. "
        b"Two different statements referencing the ligature glyph 'glyph6' "
        b"have assigned the same mark class to the same ligature component."),
    ('test_mark_ligature_conflict_pos',
        b"[ERROR] <SourceSans-Test> MarkToLigature statement error in feature "
        b"'tst1'. Glyph 'glyph6' contains a duplicate mark class assignment "
        b"for one of the ligature components."),
    ('test_mark_class_glyph_conflict',
        b"[ERROR] <SourceSans-Test> In feature 'tst1', glyph 'glyph2' "
        b"is repeated in the current class definition. Mark class: "
        b"@TOP_MARKS."),
    ('test_mark_class_glyph_conflict',
        b"[ERROR] <SourceSans-Test> In feature 'tst1', glyph 'glyph1' "
        b"occurs in two different mark classes. Previous mark class: "
        b"@TOP_MARKS. Current mark class: @BOTTOM_MARKS."),
    ('test_base_anchor_errors_pos',
        b"[ERROR] <SourceSans-Test> MarkToBase or MarkToMark error in "
        b"feature 'tst1'. Another statement has already assigned the current "
        b"mark class to another anchor point on glyph 'glyph7'"),
    ('test_base_anchor_errors_pos',
        b"[WARNING] <SourceSans-Test> MarkToBase or MarkToMark error in "
        b"feature 'tst1'. Glyph 'glyph5' does not have an anchor point "
        b"for a mark class that was used in a previous statement in the "
        b"same lookup table"),
    ('test_cursive_duplicate_glyph_pos',
        b"[ERROR] <SourceSans-Test> Cursive statement error in feature "
        b"'tst1'. A previous statement has already referenced glyph "
        b"'glyph1'."),
])
def test_overflow_report_bug313(feat_name, error_msg):
    input_filename = 'bug313/font.pfa'
    feat_filename = f'bug313/{feat_name}.fea'
    otf_path = get_temp_file_path()

    stderr_path = runner(
        CMD + ['-s', '-e', '-o', 'shw',
               'f', f'_{get_input_path(input_filename)}',
               'ff', f'_{get_input_path(feat_filename)}',
               'o', f'_{otf_path}'])

    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert error_msg in output


def test_feature_recursion_bug628():
    input_filename = 'bug628/font.pfa'
    feat_filename = 'bug628/feat.fea'
    otf_path = get_temp_file_path()

    stderr_path = runner(
        CMD + ['-s', '-e', '-o', 'shw',
               'f', f'_{get_input_path(input_filename)}',
               'ff', f'_{get_input_path(feat_filename)}',
               'o', f'_{otf_path}'])

    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert(b"[FATAL] <SourceSans-Test> Can't include [feat.fea]; maximum "
           b"include levels <50> reached") in output


@pytest.mark.parametrize('arg', [[], ['gs']])
def test_recalculate_bbox_bug617(arg):
    input_filename = "bug617/font.pfa"
    goadb_filename = "bug617/goadb.txt"
    ttx_filename = f"bug617/{'no_' if not arg else ''}gs_opt.ttx"
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'gf', f'_{get_input_path(goadb_filename)}',
                        'o', f'_{actual_path}', 'r'] + arg)
    actual_ttx = generate_ttx_dump(actual_path, ['head', 'CFF '])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value='])


def test_contextual_multiple_substitutions_bug725():
    input_filename = "bug725/font.pfa"
    feat_filename = "bug725/feat.fea"
    actual_path = get_temp_file_path()
    ttx_filename = "bug725.ttx"
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GSUB'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_notdef_in_glyph_class_bug726():
    input_filename = "bug726/font.pfa"
    feat_filename = "bug726/feat.fea"
    actual_path = get_temp_file_path()
    ttx_filename = "bug726.ttx"
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GDEF'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_overflow_bug731():
    input_filename = 'bug731/font.pfa'
    feat_filename = 'bug731/feat.fea'
    otf_path = get_temp_file_path()

    stderr_path = runner(
        CMD + ['-s', '-e', '-o', 'shw',
               'f', f'_{get_input_path(input_filename)}',
               'ff', f'_{get_input_path(feat_filename)}',
               'o', f'_{otf_path}'])

    with open(stderr_path, 'rb') as f:
        output = f.read()
    assert(b"[FATAL] <SourceSans-Test> subtable offset too large (1003c) in "
           b"lookup 0 type 3") in output


def test_base_anchor_bug811():
    input_filename = 'bug811/font.pfa'
    feat_filename = 'bug811/feat.fea'
    ttx_filename = 'bug811.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GPOS'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-s', '<ttFont sfntVersion'])


def test_max_revision_bug876():
    input_filename = 'bug876/font.pfa'
    feat_filename = 'bug876/feat.fea'
    ttx_filename = 'bug876.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['head'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value='])


def test_infinite_loop_with_dflt_lookups_bug965():
    input_filename = 'bug965/font.pfa'
    feat_filename = 'bug965/feat.fea'
    ttx_filename = 'bug965.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GSUB'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value='])


def test_bug993():
    input_filename = 'fealib/font.pfa'
    feat_filename = 'bug993/feat.fea'
    ttx_filename = 'bug993.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GPOS'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value='])


def test_bug1006():
    input_filename = 'fealib/font.pfa'
    feat_filename = 'bug1006/feat.fea'
    ttx_filename = 'bug1006.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['GPOS'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value='])


@pytest.mark.parametrize(
    'cmap_filename, cs, expectedOS2',
    [('UniCN-UTF32-H', '25', 'expected_OS2_SC.ttx'),
     ('UniJP-UTF32-H', '1', 'expected_OS2_JP.ttx'),
     ('UniKR-UTF32-H', '3', 'expected_OS2_KR.ttx'),
     ('UniTW-UTF32-H', '2', 'expected_OS2_TC.ttx')])
def test_bug1040(cmap_filename, cs, expectedOS2):
    out_path = get_temp_file_path()
    cmd = CMD + ['-o', 'f', f'_{get_input_path("bug1040/cidfont.ps")}',
                       'cs', f'_{cs}',
                       'o', f'_{out_path}',
                       'ch', f'_{get_input_path("bug1040/" + cmap_filename)}']
    runner(cmd)

    actual_ttx = generate_ttx_dump(out_path, ["OS/2"])
    expected_ttx = get_expected_path("bug1040/" + expectedOS2)

    assert differ([expected_ttx, actual_ttx, '-l', '2'])


def test_parameter_offset_overflow_bug1017():
    # Since the fix for issue 1017, it is not practical to actually cause
    # this overflow, which used to be tested as bug746.  What we do
    # instead is to verify that the FeatParams blocks are written right
    # after the FeatureList, and before the LookupList, by comparing with
    # the expected low level spot GSUB dump.
    input_filename = 'bug1017/font.pfa'
    feat_filename = 'bug1017/feat.fea'
    txt_filename = 'bug1017.GSUB.txt'
    otf_path = get_temp_file_path()

    runner(
        CMD + ['-s', '-e', '-o', 'shw',
               'f', f'_{get_input_path(input_filename)}',
               'ff', f'_{get_input_path(feat_filename)}',
               'o', f'_{otf_path}'])

    actual_txt = generate_spot_dumptables(otf_path, ['GSUB'])
    expected_txt = get_expected_path(txt_filename)
    assert differ([expected_txt, actual_txt])


def test_bug1178():
    input_filename = 'fealib/font.pfa'
    feat_filename = 'fealib/bug1178.fea'
    ttx_filename = 'fealib/bug1178.ttx'
    actual_path = get_temp_file_path()
    runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                        'ff', f'_{get_input_path(feat_filename)}',
                        'o', f'_{actual_path}'])
    actual_ttx = generate_ttx_dump(actual_path, ['name', 'STAT'])
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx,
                   '-s',
                   '<ttFont sfntVersion' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value=',
                   '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])


TEST_FEATURE_FILES = [
    "Attach", "enum", "markClass", "language_required", "GlyphClassDef",
    "LigatureCaretByIndex", "LigatureCaretByPos", "lookup", "lookupflag",
    "feature_aalt", "ignore_pos", "GPOS_1", "GPOS_1_zero", "GPOS_2", "GPOS_2b",
    "GPOS_3", "GPOS_4", "GPOS_5", "GPOS_6", "GPOS_8", "GSUB_2", "GSUB_3",
    "GSUB_6", "GSUB_8", "spec4h1", "spec4h2", "spec5d1", "spec5d2", "spec5fi1",
    "spec5fi2", "spec5fi3", "spec5fi4", "spec5f_ii_1", "spec5f_ii_2",
    "spec5f_ii_3", "spec5f_ii_4", "spec5h1", "spec6b_ii", "spec6d2", "spec6e",
    "spec6f", "spec6h_ii", "spec6h_iii_1", "spec6h_iii_3d", "spec8a", "spec8b",
    "spec8c", "spec8d", "spec9a", "spec9b", "spec9c1", "spec9c2", "spec9c3",
    "spec9d", "spec9e", "spec9f", "spec9g", "spec10", "bug453", "bug457",
    "bug463", "bug501", "bug502", "bug504", "bug505", "bug506", "bug509",
    "bug512", "bug514", "bug568", "bug633", "bug1307", "bug1459", "name",
    "size", "size2", "multiple_feature_blocks", "omitted_GlyphClassDef",
    "ZeroValue_SinglePos_horizontal", "ZeroValue_SinglePos_vertical",
    "ZeroValue_PairPos_horizontal", "ZeroValue_PairPos_vertical",
    "ZeroValue_ChainSinglePos_horizontal", "ZeroValue_ChainSinglePos_vertical",
    "PairPosSubtable", "MultipleLookupsPerGlyph", "MultipleLookupsPerGlyph2",
    "name_after_STAT.bad", "incomplete_glyph_range.bad",
]

TEST_FEATURE_FILES_XFAIL = [
    "GPOS_1",            # <device>
    "GSUB_6",            # Contextual alternate rule not yet supported
    "bug1459",           # <NULL>
    "language_required",  # required
    "PairPosSubtable",   # https://github.com/adobe-type-tools/afdko/issues/971
    "bug1307",           # https://github.com/adobe-type-tools/afdko/issues/966
    "name_after_STAT.bad",
    "incomplete_glyph_range.bad",
]

TEST_FEATURE_FILES_TABLES = {
    "name": ["name"],
    "spec8b": ["name", "GPOS"],
    "spec8c": ["name", "GSUB"],
    "spec8d": ["name", "GSUB"],
    "spec9c1": ["head"],
    "spec9c2": ["head"],
    "spec9c3": ["head"],
    "spec9d": ["hhea"],
    "spec9e": ["name"],
    "spec9f": ["OS/2"],
    "spec9g": ["vhea"],
}

TEST_TABLES = ['BASE', 'GDEF', 'GSUB', 'GPOS']


@pytest.mark.parametrize('name', TEST_FEATURE_FILES)
def test_feature_file(name):
    try:
        input_filename = "fealib/font.pfa"
        feat_filename = f"fealib/{name}.fea"
        ttx_filename = f"fealib/{name}.ttx"
        actual_path = get_temp_file_path()

        runner(CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                            'ff', f'_{get_input_path(feat_filename)}',
                            'o', f'_{actual_path}'])

        tables = TEST_FEATURE_FILES_TABLES.get(name, TEST_TABLES)
        actual_ttx = generate_ttx_dump(actual_path, tables)
        expected_ttx = get_expected_path(ttx_filename)
        assert differ([expected_ttx, actual_ttx, '-l', '2',
                       '-s',
                       '    <checkSumAdjustment value=' + SPLIT_MARKER +
                       '    <checkSumAdjustment value=' + SPLIT_MARKER +
                       '    <created value=' + SPLIT_MARKER +
                       '    <modified value=',
                       '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])
    except Exception:
        if name in TEST_FEATURE_FILES_XFAIL:
            pytest.xfail()
        raise


@pytest.mark.parametrize('path', glob.glob(get_input_path("spec/*.fea")))
def test_spec(path):
    name = os.path.splitext(os.path.basename(path))[0]
    if ".cid" in name:
        input_filename = "spec/fontcid.ps"
    else:
        input_filename = "spec/font.pfa"
    feat_filename = f"spec/{name}.fea"
    ttx_filename = f"spec/{name}.ttx"
    cmap_filename = "spec/Identity-H"
    actual_path = get_temp_file_path()

    cmd = CMD + ['-o', 'f', f'_{get_input_path(input_filename)}',
                       'ff', f'_{get_input_path(feat_filename)}',
                       'o', f'_{actual_path}',
                       'ch', f'_{get_input_path(cmap_filename)}']

    if name.endswith(".bad"):
        with pytest.raises(subprocess.CalledProcessError):
            runner(cmd)
        return
    runner(cmd)

    tables = TEST_TABLES
    with open(get_input_path(feat_filename)) as fp:
        line = fp.readline()
        if line.startswith("#") and "TABLES" in line:
            tables = line.split("# TABLES:")[1].strip().split(",")

    actual_ttx = generate_ttx_dump(actual_path, tables)
    expected_ttx = get_expected_path(ttx_filename)
    assert differ([expected_ttx, actual_ttx, '-l', '2',
                   '-s',
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <checkSumAdjustment value=' + SPLIT_MARKER +
                   '    <created value=' + SPLIT_MARKER +
                   '    <modified value=',
                   '-r', r'^\s+Version.*;hotconv.*;makeotfexe'])



import pytest
import subprocess

from runner import main as runner
from differ import main as differ
from test_utils import get_expected_path

TOOL = 'detype1'


# -----
# Tests
# -----

@pytest.mark.parametrize('arg', ['-h'])
def test_exit_known_option(arg):
    assert subprocess.call([TOOL, arg]) == 0


@pytest.mark.parametrize('arg', ['-v', '-u'])
def test_exit_unknown_option(arg):
    assert subprocess.call([TOOL, arg]) == 1


def test_run_on_pfa_data():
    actual_path = runner(['-t', TOOL, '-s', '-f', 'type1.pfa'])
    expected_path = get_expected_path('type1.txt')
    assert differ([expected_path, actual_path])
